"""
@file
@brief Validates runtime for many :epkg:`scikit-learn` operators.
The submodule relies on :epkg:`onnxconverter_common`,
:epkg:`sklearn-onnx`.
"""
import math
import copy
from timeit import Timer
import os
import warnings
from importlib import import_module
import pickle
from time import perf_counter
import numpy
from sklearn.base import BaseEstimator
from sklearn.linear_model._base import LinearModel
from sklearn.model_selection import train_test_split
from sklearn import __all__ as sklearn__all__, __version__ as sklearn_version
from .validate_problems import _problems


class RuntimeBadResultsError(RuntimeError):
    """
    Raised when the results are too different from
    :epkg:`scikit-learn`.
    """

    def __init__(self, msg, obs):
        """
        :param msg: to display
        :param obs: observations
        """
        RuntimeError.__init__(self, msg)
        self.obs = obs


def _dictionary2str(di):
    el = []
    for k in sorted(di):
        el.append('{}={}'.format(k, di[k]))
    return '/'.join(el)


def modules_list():
    """
    Returns modules and versions currently used.

    .. runpython::
        :showcode:
        :rst:
        :warningout: DeprecationWarning

        from mlprodict.onnxrt.validate.validate_helper import modules_list
        from pyquickhelper.pandashelper import df2rst
        from pandas import DataFrame
        print(df2rst(DataFrame(modules_list())))
    """
    def try_import(name):
        try:
            mod = import_module(name)
        except ImportError:  # pragma: no cover
            return None
        return (dict(name=name, version=mod.__version__)
                if hasattr(mod, '__version__') else dict(name=name))

    rows = []
    for name in sorted(['pandas', 'numpy', 'sklearn', 'mlprodict',
                        'skl2onnx', 'onnxmltools', 'onnx', 'onnxruntime',
                        'scipy']):
        res = try_import(name)
        if res is not None:
            rows.append(res)
    return rows


def _dispsimple(arr, fLOG):
    if isinstance(arr, (tuple, list)):
        for i, a in enumerate(arr):
            fLOG("output %d" % i)
            _dispsimple(a, fLOG)
    elif hasattr(arr, 'shape'):
        if len(arr.shape) == 1:
            threshold = 8
        else:
            threshold = min(
                50, min(50 // arr.shape[1], 8) * arr.shape[1])
        fLOG(numpy.array2string(arr, max_line_width=120,
                                suppress_small=True,
                                threshold=threshold))
    else:  # pragma: no cover
        s = str(arr)
        if len(s) > 50:
            s = s[:50] + "..."
        fLOG(s)


def _merge_options(all_conv_options, aoptions):
    if aoptions is None:
        return copy.deepcopy(all_conv_options)
    if not isinstance(aoptions, dict):
        return copy.deepcopy(aoptions)  # pragma: no cover
    merged = {}
    for k, v in all_conv_options.items():
        if k in aoptions:
            merged[k] = _merge_options(v, aoptions[k])
        else:
            merged[k] = copy.deepcopy(v)
    for k, v in aoptions.items():
        if k in all_conv_options:
            continue
        merged[k] = copy.deepcopy(v)
    return merged


def sklearn_operators(subfolder=None, extended=False,
                      experimental=True):
    """
    Builds the list of operators from :epkg:`scikit-learn`.
    The function goes through the list of submodule
    and get the list of class which inherit from
    :epkg:`scikit-learn:base:BaseEstimator`.

    :param subfolder: look into only one subfolder
    :param extended: extends the list to the list of operators
        this package implements a converter for
    :param experimental: includes experimental module from
        :epkg:`scikit-learn` (see `sklearn.experimental
        <https://github.com/scikit-learn/scikit-learn/
        tree/master/sklearn/experimental>`_)
    :return: the list of found operators
    """
    if experimental:
        from sklearn.experimental import (  # pylint: disable=W0611
            enable_hist_gradient_boosting,
            enable_iterative_imputer)

    subfolders = sklearn__all__ + ['mlprodict.onnx_conv']
    found = []
    for subm in sorted(subfolders):
        if isinstance(subm, list):
            continue  # pragma: no cover
        if subfolder is not None and subm != subfolder:
            continue

        if subm == 'feature_extraction':
            subs = [subm, 'feature_extraction.text']
        else:
            subs = [subm]

        for sub in subs:
            if '.' in sub and sub not in {'feature_extraction.text'}:
                name_sub = sub
            else:
                name_sub = "{0}.{1}".format("sklearn", sub)
            try:
                mod = import_module(name_sub)
            except ModuleNotFoundError:
                continue

            if hasattr(mod, "register_converters"):
                fct = getattr(mod, "register_converters")
                cls = fct()
            else:
                cls = getattr(mod, "__all__", None)
                if cls is None:
                    cls = list(mod.__dict__)
                cls = [mod.__dict__[cl] for cl in cls]

            for cl in cls:
                try:
                    issub = issubclass(cl, BaseEstimator)
                except TypeError:
                    continue
                if cl.__name__ in {'Pipeline', 'ColumnTransformer',
                                   'FeatureUnion', 'BaseEstimator',
                                   'BaseEnsemble', 'BaseDecisionTree'}:
                    continue
                if cl.__name__ in {'CustomScorerTransform'}:
                    continue
                if (sub in {'calibration', 'dummy', 'manifold'} and
                        'Calibrated' not in cl.__name__):
                    continue
                if issub:
                    pack = "sklearn" if sub in sklearn__all__ else cl.__module__.split('.')[
                        0]
                    found.append(
                        dict(name=cl.__name__, subfolder=sub, cl=cl, package=pack))

    if extended:
        from ...onnx_conv import register_converters
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ResourceWarning)
            models = register_converters(True)

        done = set(_['name'] for _ in found)
        for m in models:
            try:
                name = m.__module__.split('.')
            except AttributeError as e:  # pragma: no cover
                raise AttributeError("Unexpected value, m={}".format(m)) from e
            sub = '.'.join(name[1:])
            pack = name[0]
            if m.__name__ not in done:
                found.append(
                    dict(name=m.__name__, cl=m, package=pack, sub=sub))

    # let's remove models which cannot predict
    all_found = found
    found = []
    for mod in all_found:
        cl = mod['cl']
        if hasattr(cl, 'fit_predict') and not hasattr(cl, 'predict'):
            continue
        if hasattr(cl, 'fit_transform') and not hasattr(cl, 'transform'):
            continue
        if (not hasattr(cl, 'transform') and
                not hasattr(cl, 'predict') and
                not hasattr(cl, 'decision_function')):
            continue
        found.append(mod)
    return found


def _measure_time(fct, repeat=1, number=1, first_run=True):
    """
    Measures the execution time for a function.

    :param fct: function to measure
    :param repeat: number of times to repeat
    :param number: number of times between two measures
    :param first_run: if True, runs the function once before measuring
    :return: last result, average, values
    """
    res = None
    values = []
    if first_run:
        fct()
    for __ in range(repeat):
        begin = perf_counter()
        for _ in range(number):
            res = fct()
        end = perf_counter()
        values.append(end - begin)
    if repeat * number == 1:
        return res, values[0], values
    return res, sum(values) / (repeat * number), values  # pragma: no cover


def _shape_exc(obj):
    if hasattr(obj, 'shape'):
        return obj.shape
    if isinstance(obj, (list, dict, tuple)):
        return "[{%d}]" % len(obj)
    return None


def dump_into_folder(dump_folder, obs_op=None, is_error=True,
                     **kwargs):
    """
    Dumps information when an error was detected
    using :epkg:`*py:pickle`.

   :param dump_folder: dump_folder
   :param obs_op: obs_op (information)
   :param is_error: is it an error or not?
   :param kwargs: additional parameters
    :return: name
    """
    if dump_folder is None:
        raise ValueError("dump_folder cannot be None.")
    optim = obs_op.get('optim', '')
    optim = str(optim)
    optim = optim.replace("<class 'sklearn.", "")
    optim = optim.replace("<class '", "")
    optim = optim.replace(" ", "")
    optim = optim.replace(">", "")
    optim = optim.replace("=", "")
    optim = optim.replace("{", "")
    optim = optim.replace("}", "")
    optim = optim.replace(":", "")
    optim = optim.replace("'", "")
    optim = optim.replace("/", "")
    optim = optim.replace("\\", "")
    parts = (obs_op['runtime'], obs_op['name'], obs_op['scenario'],
             obs_op['problem'], optim,
             "op" + str(obs_op.get('opset', '-')),
             "nf" + str(obs_op.get('n_features', '-')))
    name = "dump-{}-{}.pkl".format(
        "ERROR" if is_error else "i",
        "-".join(map(str, parts)))
    name = os.path.join(dump_folder, name)
    obs_op = obs_op.copy()
    fcts = [k for k in obs_op if k.startswith('lambda')]
    for fct in fcts:
        del obs_op[fct]
    kwargs.update({'obs_op': obs_op})
    with open(name, "wb") as f:
        pickle.dump(kwargs, f)
    return name


def default_time_kwargs():
    """
    Returns default values *number* and *repeat* to measure
    the execution of a function.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.onnxrt.validate.validate_helper import default_time_kwargs
        import pprint
        pprint.pprint(default_time_kwargs())

    keys define the number of rows,
    values defines *number* and *repeat*.
    """
    return {
        1: dict(number=30, repeat=20),
        10: dict(number=20, repeat=20),
        100: dict(number=8, repeat=10),
        1000: dict(number=5, repeat=5),
        10000: dict(number=3, repeat=3),
    }


def measure_time(stmt, x, repeat=10, number=50, div_by_number=False, first_run=True):
    """
    Measures a statement and returns the results as a dictionary.

    :param stmt: string
    :param x: matrix
    :param repeat: average over *repeat* experiment
    :param number: number of executions in one row
    :param div_by_number: divide by the number of executions
    :param first_run: if True, runs the function once before measuring
    :return: dictionary

    See `Timer.repeat <https://docs.python.org/3/library/timeit.html?timeit.Timer.repeat>`_
    for a better understanding of parameter *repeat* and *number*.
    The function returns a duration corresponding to
    *number* times the execution of the main statement.
    """
    if x is None:
        raise ValueError("x cannot be None")  # pragma: no cover

    try:
        stmt(x)
    except RuntimeError as e:  # pragma: no cover
        raise RuntimeError("{}-{}".format(type(x), x.dtype)) from e

    def fct():
        stmt(x)

    if first_run:
        fct()
    tim = Timer(fct)
    res = numpy.array(tim.repeat(repeat=repeat, number=number))
    total = numpy.sum(res)
    if div_by_number:
        res /= number
    mean = numpy.mean(res)
    dev = numpy.mean(res ** 2)
    dev = max(0, (dev - mean**2)) ** 0.5
    mes = dict(average=mean, deviation=dev, min_exec=numpy.min(res),
               max_exec=numpy.max(res), repeat=repeat, number=number,
               total=total)
    return mes


def _multiply_time_kwargs(time_kwargs, time_kwargs_fact, inst):
    """
    Multiplies values in *time_kwargs* following strategy
    *time_kwargs_fact* for a given model *inst*.

    :param time_kwargs: see below
    :param time_kwargs_fact: see below
    :param inst: :epkg:`scikit-learn` model
    :return: new *time_kwargs*

    Possible values for *time_kwargs_fact*:

    - a integer: multiplies *number* by this number
    - `'lin'`: multiplies value *number* for linear models depending
      on the number of rows to process (:math:`\\propto 1/\\log_{10}(n)`)

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from pprint import pprint
        from sklearn.linear_model import LinearRegression
        from mlprodict.onnxrt.validate.validate_helper import (
            default_time_kwargs, _multiply_time_kwargs)

        lr = LinearRegression()
        kw = default_time_kwargs()
        pprint(kw)

        kw2 = _multiply_time_kwargs(kw, 'lin', lr)
        pprint(kw2)
    """
    if time_kwargs is None:
        raise ValueError("time_kwargs cannot be None.")  # pragma: no cover
    if time_kwargs_fact in ('', None):
        return time_kwargs
    try:
        vi = int(time_kwargs_fact)
        time_kwargs_fact = vi
    except (TypeError, ValueError):
        pass
    if isinstance(time_kwargs_fact, int):
        time_kwargs_modified = copy.deepcopy(time_kwargs)
        for k in time_kwargs_modified:
            time_kwargs_modified[k]['number'] *= time_kwargs_fact
        return time_kwargs_modified
    if time_kwargs_fact == 'lin':
        if isinstance(inst, LinearModel):
            time_kwargs_modified = copy.deepcopy(time_kwargs)
            for k in time_kwargs_modified:
                kl = max(int(math.log(k) / math.log(10) + 1e-5), 1)
                f = max(int(10 / kl + 0.5), 1)
                time_kwargs_modified[k]['number'] *= f
                time_kwargs_modified[k]['repeat'] *= 1
            return time_kwargs_modified
        return time_kwargs
    raise ValueError(  # pragma: no cover
        "Unable to interpret time_kwargs_fact='{}'.".format(
            time_kwargs_fact))


def _get_problem_data(prob, n_features):
    data_problem = _problems[prob](n_features=n_features)
    if len(data_problem) == 6:
        X_, y_, init_types, method, output_index, Xort_ = data_problem
        dofit = True
    elif len(data_problem) == 7:
        X_, y_, init_types, method, output_index, Xort_, dofit = data_problem
    else:
        raise RuntimeError(  # pragma: no cover
            "Unable to interpret problem '{}'.".format(prob))
    if (len(X_.shape) == 2 and X_.shape[1] != n_features and
            n_features is not None):
        raise RuntimeError(  # pragma: no cover
            "Problem '{}' with n_features={} returned {} features"
            "(func={}).".format(prob, n_features, X_.shape[1],
                                _problems[prob]))
    if y_ is None:
        (X_train, X_test, Xort_train,  # pylint: disable=W0612
            Xort_test) = train_test_split(
                X_, Xort_, random_state=42)
        y_train, y_test = None, None
    else:
        (X_train, X_test, y_train, y_test,  # pylint: disable=W0612
            Xort_train, Xort_test) = train_test_split(
                X_, y_, Xort_, random_state=42)
    if isinstance(init_types, tuple):
        init_types, conv_options = init_types
    else:
        conv_options = None

    if isinstance(method, tuple):
        method_name, predict_kwargs = method
    else:
        method_name = method
        predict_kwargs = {}

    return (X_train, X_test, y_train,
            y_test, Xort_test,
            init_types, conv_options, method_name,
            output_index, dofit, predict_kwargs)



# -*- encoding: utf-8 -*-
"""
@file
@brief Defines new operators.
"""


class OperatorSchema:
    """
    Defines a schema for operators added in this package
    such as @see cl TreeEnsembleRegressorDouble.
    """

    def __init__(self, name):
        self.name = name
        self.domain = 'mlprodict'



"""
@file
@brief Shortcuts to tools.
"""

from .asv_options_helper import get_opset_number_from_onnx, get_ir_version_from_onnx
from .code_helper import change_style
from .speed_measure import measure_time



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.defs import onnx_opset_version
from ._op import OpRun
from ..shape_object import ShapeObject


class CommonRNN(OpRun):

    def __init__(self, onnx_node, expected_attributes=None, desc=None,
                 **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=expected_attributes,
                       **options)

        if self.direction in ("forward", "reverse"):
            self.num_directions = 1
        elif self.direction == "bidirectional":
            self.num_directions = 2
        else:
            raise RuntimeError(  # pragma: no cover
                "Unknown direction '{}'.".format(self.direction))

        if len(self.activation_alpha) != self.num_directions:
            raise RuntimeError(  # pragma: no cover
                "activation_alpha must have the same size as num_directions={}".format(
                    self.num_directions))
        if len(self.activation_beta) != self.num_directions:
            raise RuntimeError(  # pragma: no cover
                "activation_beta must have the same size as num_directions={}".format(
                    self.num_directions))

        self.f1 = self.choose_act(self.activations[0],
                                  self.activation_alpha[0],
                                  self.activation_beta[0])
        if len(self.activations) > 1:
            self.f2 = self.choose_act(self.activations[1],
                                      self.activation_alpha[1],
                                      self.activation_beta[1])
        self.nb_outputs = len(onnx_node.output)
        if getattr(self, 'layout', 0) != 0:
            raise NotImplementedError(
                "The runtime is not implemented when layout=%r != 0." % self.layout)

    def choose_act(self, name, alpha, beta):
        if name == b"Tanh":
            return self._f_tanh
        if name == b"Affine":
            return lambda x: x * alpha + beta
        raise RuntimeError(  # pragma: no cover
            "Unknown activation function '{}'.".format(name))

    def _f_tanh(self, x):
        return numpy.tanh(x)

    def _step(self, X, R, B, W, H_0):
        h_list = []
        H_t = H_0
        for x in numpy.split(X, X.shape[0], axis=0):
            H = self.f1(numpy.dot(x, numpy.transpose(W)) +
                        numpy.dot(H_t, numpy.transpose(R)) +
                        numpy.add(*numpy.split(B, 2)))
            h_list.append(H)
            H_t = H
        concatenated = numpy.concatenate(h_list)
        if self.num_directions == 1:
            output = numpy.expand_dims(concatenated, 1)
        return output, h_list[-1]

    def _run(self, X, W, R, B=None, sequence_lens=None, initial_h=None):  # pylint: disable=W0221
        self.num_directions = W.shape[0]

        if self.num_directions == 1:
            R = numpy.squeeze(R, axis=0)
            W = numpy.squeeze(W, axis=0)
            if B is not None:
                B = numpy.squeeze(B, axis=0)
            if sequence_lens is not None:
                sequence_lens = numpy.squeeze(sequence_lens, axis=0)
            if initial_h is not None:
                initial_h = numpy.squeeze(initial_h, axis=0)

            hidden_size = R.shape[-1]
            batch_size = X.shape[1]

            b = (B if B is not None else
                 numpy.zeros(2 * hidden_size, dtype=numpy.float32))
            h_0 = (initial_h if initial_h is not None else
                   numpy.zeros((batch_size, hidden_size), dtype=numpy.float32))

            B = b
            H_0 = h_0
        else:
            raise NotImplementedError()  # pragma: no cover

        Y, Y_h = self._step(X, R, B, W, H_0)
        return (Y, ) if self.nb_outputs == 1 else (Y, Y_h)

    def _infer_shapes(self, X, W, R, B=None, sequence_lens=None, initial_h=None):  # pylint: disable=W0221
        num_directions = W.shape[0]

        if num_directions == 1:
            hidden_size = R[-1]
            batch_size = X[1]
            y_shape = ShapeObject((X[0], num_directions, batch_size, hidden_size),
                                  dtype=X.dtype)
        else:
            raise NotImplementedError()  # pragma: no cover
        if self.nb_outputs == 1:
            return (y_shape, )
        y_h_shape = ShapeObject((num_directions, batch_size, hidden_size),
                                dtype=X.dtype)
        return (y_shape, y_h_shape)

    def _infer_types(self, X, W, R, B=None, sequence_lens=None, initial_h=None):  # pylint: disable=W0221
        return (X, X)


class RNN_7(CommonRNN):

    atts = {
        'activation_alpha': [0.],
        'activation_beta': [0.],
        'activations': ['tanh', 'tanh'],
        'clip': [],
        'direction': 'forward',
        'hidden_size': None,
    }

    def __init__(self, onnx_node, desc=None, **options):
        CommonRNN.__init__(self, onnx_node, desc=desc,
                           expected_attributes=RNN_7.atts,
                           **options)


class RNN_14(CommonRNN):

    atts = {
        'activation_alpha': [0.],
        'activation_beta': [0.],
        'activations': ['tanh', 'tanh'],
        'clip': [],
        'direction': 'forward',
        'hidden_size': None,
        'layout': 0,
    }

    def __init__(self, onnx_node, desc=None, **options):
        CommonRNN.__init__(self, onnx_node, desc=desc,
                           expected_attributes=RNN_14.atts,
                           **options)


if onnx_opset_version() >= 14:
    RNN = RNN_14
else:  # pragma: no cover
    RNN = RNN_7



"""
@file
@brief Common class for classifiers supporting strings.
"""
import numpy


class _ClassifierCommon:
    """
    Labels strings are not natively implemented in C++ runtime.
    The class stores the strings labels, replaces them by
    integer, calls the C++ codes and then replaces them by strings.
    """

    def _post_process_label_attributes(self):
        """
        Replaces string labels by int64 labels.
        It creates attributes *_classlabels_int64s_string*.
        """
        name_int = 'classlabels_int64s' if hasattr(
            self, 'classlabels_int64s') else 'classlabels_ints'
        if (hasattr(self, 'classlabels_strings') and
                len(self.classlabels_strings) > 0):  # pylint: disable=E0203
            if hasattr(self, name_int) and len(getattr(self, name_int)) != 0:
                raise RuntimeError(  # pragma: no cover
                    "'%s' must be empty if "
                    "'classlabels_strings' is not." % name_int)
            setattr(self, name_int, numpy.arange(len(self.classlabels_strings),  # pylint: disable=E0203
                                                 dtype=numpy.int64))
            self._classlabels_int64s_string = self.classlabels_strings  # pylint: disable=E0203
            self.classlabels_strings = numpy.empty(
                shape=(0, ), dtype=numpy.str_)
        else:
            self._classlabels_int64s_string = None

    def _post_process_predicted_label(self, label, scores):
        """
        Replaces int64 predicted labels by the corresponding
        strings.
        """
        if self._classlabels_int64s_string is not None:
            label = numpy.array(
                [self._classlabels_int64s_string[i] for i in label])
        return label, scores



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from scipy.sparse import coo_matrix
from ._op import OpRun, RuntimeTypeError
from ..shape_object import ShapeObject


class DictVectorizer(OpRun):

    atts = {'int64_vocabulary': numpy.empty(0, dtype=numpy.int64),
            'string_vocabulary': numpy.empty(0, dtype=numpy.str_)}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=DictVectorizer.atts,
                       **options)
        self.dict_labels = {}
        if len(self.int64_vocabulary) > 0:
            for i, v in enumerate(self.int64_vocabulary):
                self.dict_labels[v] = i
            self.is_int = True
        else:
            for i, v in enumerate(self.string_vocabulary):
                self.dict_labels[v.decode('utf-8')] = i
            self.is_int = False
        if len(self.dict_labels) == 0:
            raise RuntimeError(  # pragma: no cover
                "int64_vocabulary and string_vocabulary cannot be both empty.")

    def _run(self, x):  # pylint: disable=W0221
        if not isinstance(x, (numpy.ndarray, list)):
            raise RuntimeTypeError(  # pragma: no cover
                "x must be iterable not {}.".format(type(x)))
        values = []
        rows = []
        cols = []
        for i, row in enumerate(x):
            for k, v in row.items():
                values.append(v)
                rows.append(i)
                cols.append(self.dict_labels[k])
        values = numpy.array(values)
        rows = numpy.array(rows)
        cols = numpy.array(cols)
        return (coo_matrix((values, (rows, cols)), shape=(len(x), len(self.dict_labels))), )

    def _infer_shapes(self, x):  # pylint: disable=W0221
        pref = str(hex(id(self))[2:])
        return (ShapeObject(["ndv%s_0" % pref, "N%s_1" % pref], dtype=x.dtype), )

    def _infer_types(self, x):  # pylint: disable=W0221
        return (x, )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ..shape_object import ShapeObject


class ConstantOfShape(OpRun):

    atts = {'value': numpy.array([0], dtype=numpy.float32)}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=ConstantOfShape.atts,
                       **options)
        self.cst = (self.value[0]
                    if isinstance(self.value, numpy.ndarray)
                    else self.value)
        if not isinstance(self.cst, (float, numpy.float32, numpy.float64,
                                     numpy.int64, numpy.int32, numpy.bool_,
                                     numpy.float16)):
            raise TypeError(  # pragma: no cover
                "cst must be a real not {}".format(type(self.cst)))

    def _run(self, data):  # pylint: disable=W0221
        res = numpy.full(tuple(data), self.cst)
        return (res, )

    def _infer_shapes(self, data):  # pylint: disable=W0221
        # pref = str(hex(id(self))[2:])
        return (ShapeObject(None, self.cst.dtype), )

    def _infer_types(self, data):  # pylint: disable=W0221
        # pref = str(hex(id(self))[2:])
        if isinstance(self.cst, numpy.ndarray):
            return (self.cst.dtype, )
        return (type(self.cst), )

    def _infer_sizes(self, *args, **kwargs):
        res = self.run(*args, **kwargs)
        return (dict(temp=0), ) + res

    def to_python(self, inputs):
        lines = ['cst = value[0] if isinstance(value, numpy.ndarray) else value',
                 'return numpy.full(tuple(%s), cst)' % inputs[0]]
        return ("import numpy", "\n".join(lines))



"""
@file
@brief Command line to check einsum scenarios.
"""
import os


def einsum_test(equation="abc,cd->abd", shape="30", perm=False,
                runtime='python', verbose=1, fLOG=print,
                output=None, number=5, repeat=5):
    """
    Investigates whether or not the decomposing einsum is faster.

    :param equation: einsum equation to test
    :param shape: an integer (all dimension gets the same size) or
        a list of shapes in a string separated with `;`) or
        a list of integer to try out multiple shapes,
        example: `5`, `(5,5,5),(5,5)`, `5,6`
    :param perm: check on permutation or all letter permutations
    :param runtime: `'numpy'`, `'python'`, `'onnxruntime'`
    :param verbose: verbose
    :param fLOG: logging function
    :param output: output file (usually a csv file or an excel file),
        it requires pandas
    :param number: usual parameter to measure a function
    :param repeat: usual parameter to measure a function

    .. cmdref::
        :title: Investigates whether or not the decomposing einsum is faster.
        :cmd: -m mlprodict einsum_test --help
        :lid: l-cmd-einsum_test

        The command checks whether or not decomposing an einsum function
        is faster than einsum implementation.

        Example::

            python -m mlprodict einsum_test --equation="abc,cd->abd" --output=res.csv
    """
    from ..testing.einsum.einsum_bench import einsum_benchmark  # pylint: disable=E0402

    perm = perm in ('True', '1', 1, True)
    if "(" not in shape:
        if "," in shape:
            shape = list(map(int, shape.split(",")))
        else:
            shape = int(shape)
    else:
        shapes = shape.replace('(', '').replace(')', '').split(";")
        shape = []
        for sh in shapes:
            spl = sh.split(',')
            shape.append(tuple(map(int, spl)))
    verbose = int(verbose)
    number = int(number)
    repeat = int(repeat)

    res = einsum_benchmark(equation=equation, shape=shape, perm=perm,
                           runtime=runtime, use_tqdm=verbose > 0,
                           number=number, repeat=repeat)
    if output not in ('', None):
        import pandas
        df = pandas.DataFrame(res)
        ext = os.path.splitext(output)[-1]
        if ext == '.csv':
            df.to_csv(output, index=False)
            fLOG('[einsum_test] wrote file %r.' % output)
        elif ext == '.xlsx':
            df.to_excel(output, index=False)
            fLOG('[einsum_test] wrote file %r.' % output)
        else:
            raise ValueError(
                "Unknown extension %r in file %r." % (ext, output))
    else:
        for r in res:
            fLOG(r)



"""
@file
@brief Helpers to use numpy API to easily write converters
for :epkg:`scikit-learn` classes for :epkg:`onnx`.

.. versionadded:: 0.6
"""
import numpy
from sklearn.base import (
    ClassifierMixin, ClusterMixin,
    RegressorMixin, TransformerMixin)
from skl2onnx import update_registered_converter
from skl2onnx.common.data_types import Int64TensorType
from skl2onnx.algebra.onnx_ops import OnnxIdentity  # pylint: disable=E0611
from .onnx_variable import OnnxVar, TupleOnnxAny
from .onnx_numpy_wrapper import _created_classes_inst, wrapper_onnxnumpy_np
from .onnx_numpy_annotation import NDArraySameType, NDArrayType


def _common_shape_calculator_t(operator):
    if not hasattr(operator, 'onnx_numpy_fct_'):
        raise AttributeError(
            "operator must have attribute 'onnx_numpy_fct_'.")
    X = operator.inputs
    if len(X) != 1:
        raise RuntimeError(
            "This function only supports one input not %r." % len(X))
    if len(operator.outputs) != 1:
        raise RuntimeError(
            "This function only supports one output not %r." % len(
                operator.outputs))
    op = operator.raw_operator
    cl = X[0].type.__class__
    dim = [X[0].type.shape[0], getattr(op, 'n_outputs_', None)]
    operator.outputs[0].type = cl(dim)


def _shape_calculator_transformer(operator):
    """
    Default shape calculator for a transformer with one input
    and one output of the same type.

    .. versionadded:: 0.6
    """
    _common_shape_calculator_t(operator)


def _shape_calculator_regressor(operator):
    """
    Default shape calculator for a regressor with one input
    and one output of the same type.

    .. versionadded:: 0.6
    """
    _common_shape_calculator_t(operator)


def _common_shape_calculator_int_t(operator):
    if not hasattr(operator, 'onnx_numpy_fct_'):
        raise AttributeError(
            "operator must have attribute 'onnx_numpy_fct_'.")
    X = operator.inputs
    if len(X) != 1:
        raise RuntimeError(
            "This function only supports one input not %r." % len(X))
    if len(operator.outputs) != 2:
        raise RuntimeError(
            "This function only supports two outputs not %r." % len(
                operator.outputs))
    op = operator.raw_operator
    cl = X[0].type.__class__
    dim = [X[0].type.shape[0], getattr(op, 'n_outputs_', None)]
    operator.outputs[0].type = Int64TensorType(dim[:1])
    operator.outputs[1].type = cl(dim)


def _shape_calculator_classifier(operator):
    """
    Default shape calculator for a classifier with one input
    and two outputs, label (int64) and probabilites of the same type.

    .. versionadded:: 0.6
    """
    _common_shape_calculator_int_t(operator)


def _shape_calculator_cluster(operator):
    """
    Default shape calculator for a clustering with one input
    and two outputs, label (int64) and distances of the same type.

    .. versionadded:: 0.6
    """
    _common_shape_calculator_int_t(operator)


def _common_converter_t(scope, operator, container):
    if not hasattr(operator, 'onnx_numpy_fct_'):
        raise AttributeError(
            "operator must have attribute 'onnx_numpy_fct_'.")
    X = operator.inputs
    if len(X) != 1:
        raise RuntimeError(
            "This function only supports one input not %r." % len(X))
    if len(operator.outputs) != 1:
        raise RuntimeError(
            "This function only supports one output not %r." % len(
                operator.outputs))

    xvar = OnnxVar(X[0])
    fct_cl = operator.onnx_numpy_fct_

    opv = container.target_opset
    inst = fct_cl.fct(xvar, op_=operator.raw_operator)
    onx = inst.to_algebra(op_version=opv)
    final = OnnxIdentity(onx, op_version=opv,
                         output_names=[operator.outputs[0].full_name])
    final.add_to(scope, container)


def _converter_transformer(scope, operator, container):
    """
    Default converter for a transformer with one input
    and one output of the same type. It assumes instance *operator*
    has an attribute *onnx_numpy_fct_* from a function
    wrapped with decorator :func:`onnxsklearn_transformer
    <mlprodict.npy.onnx_sklearn_wrapper.onnxsklearn_transformer>`.

    .. versionadded:: 0.6
    """
    _common_converter_t(scope, operator, container)


def _converter_regressor(scope, operator, container):
    """
    Default converter for a regressor with one input
    and one output of the same type. It assumes instance *operator*
    has an attribute *onnx_numpy_fct_* from a function
    wrapped with decorator :func:`onnxsklearn_regressor
    <mlprodict.npy.onnx_sklearn_wrapper.onnxsklearn_regressor>`.

    .. versionadded:: 0.6
    """
    _common_converter_t(scope, operator, container)


def _common_converter_int_t(scope, operator, container):
    if not hasattr(operator, 'onnx_numpy_fct_'):
        raise AttributeError(
            "operator must have attribute 'onnx_numpy_fct_'.")
    X = operator.inputs
    if len(X) != 1:
        raise RuntimeError(
            "This function only supports one input not %r." % len(X))
    if len(operator.outputs) != 2:
        raise RuntimeError(
            "This function only supports two outputs not %r." % len(
                operator.outputs))

    xvar = OnnxVar(X[0])
    fct_cl = operator.onnx_numpy_fct_

    opv = container.target_opset
    inst = fct_cl.fct(xvar, op_=operator.raw_operator)
    onx = inst.to_algebra(op_version=opv)
    if isinstance(onx, TupleOnnxAny):
        if len(operator.outputs) != len(onx):
            raise RuntimeError(  # pragma: no cover
                "Mismatched number of outputs expected %d, got %d." % (
                    len(operator.outputs), len(onx)))
        for out, ox in zip(operator.outputs, onx):
            if not hasattr(ox, 'add_to'):
                raise TypeError(  # pragma: no cover
                    "Unexpected type for onnx graph %r, inst=%r." % (
                        type(ox), type(inst)))
            final = OnnxIdentity(ox, op_version=opv,
                                 output_names=[out.full_name])
            final.add_to(scope, container)
    else:
        final = OnnxIdentity(onx, op_version=opv,
                             output_names=[operator.outputs[0].full_name])
        final.add_to(scope, container)


def _converter_classifier(scope, operator, container):
    """
    Default converter for a classifier with one input
    and two outputs, label and probabilities of the same input type.
    It assumes instance *operator*
    has an attribute *onnx_numpy_fct_* from a function
    wrapped with decorator :func:`onnxsklearn_classifier
    <mlprodict.npy.onnx_sklearn_wrapper.onnxsklearn_classifier>`.

    .. versionadded:: 0.6
    """
    _common_converter_int_t(scope, operator, container)


def _converter_cluster(scope, operator, container):
    """
    Default converter for a clustering with one input
    and two outputs, label and distances of the same input type.
    It assumes instance *operator*
    has an attribute *onnx_numpy_fct_* from a function
    wrapped with decorator :func:`onnxsklearn_cluster
    <mlprodict.npy.onnx_sklearn_wrapper.onnxsklearn_cluster>`.

    .. versionadded:: 0.6
    """
    _common_converter_int_t(scope, operator, container)


_default_cvt = {
    ClassifierMixin: (_shape_calculator_classifier, _converter_classifier),
    ClusterMixin: (_shape_calculator_cluster, _converter_cluster),
    RegressorMixin: (_shape_calculator_regressor, _converter_regressor),
    TransformerMixin: (_shape_calculator_transformer, _converter_transformer),
}


def update_registered_converter_npy(
        model, alias, convert_fct, shape_fct=None, overwrite=True,
        parser=None, options=None):
    """
    Registers or updates a converter for a new model so that
    it can be converted when inserted in a *scikit-learn* pipeline.
    This function assumes the converter is written as a function
    decoarated with :func:`onnxsklearn_transformer
    <mlprodict.npy.onnx_sklearn_wrapper.onnxsklearn_transformer>`.

    :param model: model class
    :param alias: alias used to register the model
    :param shape_fct: function which checks or modifies the expected
        outputs, this function should be fast so that the whole graph
        can be computed followed by the conversion of each model,
        parallelized or not
    :param convert_fct: function which converts a model
    :param overwrite: False to raise exception if a converter
        already exists
    :param parser: overwrites the parser as well if not empty
    :param options: registered options for this converter

    The alias is usually the library name followed by the model name.

    .. versionadded:: 0.6
    """
    if (hasattr(convert_fct, "compiled") or
            hasattr(convert_fct, 'signed_compiled')):
        # type is wrapper_onnxnumpy or wrapper_onnxnumpy_np
        obj = convert_fct
    else:
        raise AttributeError(  # pragma: no cover
            "Class %r must have attribute 'compiled' or 'signed_compiled' "
            "(object=%r)." % (type(convert_fct), convert_fct))

    def addattr(operator, obj):
        operator.onnx_numpy_fct_ = obj
        return operator

    if issubclass(model, TransformerMixin):
        defcl = TransformerMixin
    elif issubclass(model, RegressorMixin):
        defcl = RegressorMixin
    elif issubclass(model, ClassifierMixin):
        defcl = ClassifierMixin
    elif issubclass(model, ClusterMixin):
        defcl = ClusterMixin
    else:
        defcl = None

    if shape_fct is not None:
        raise NotImplementedError(  # pragma: no cover
            "Custom shape calculator are not implemented yet.")

    shc = _default_cvt[defcl][0]
    local_shape_fct = (
        lambda operator: shc(addattr(operator, obj)))

    cvtc = _default_cvt[defcl][1]
    local_convert_fct = (
        lambda scope, operator, container:
        cvtc(scope, addattr(operator, obj), container))

    update_registered_converter(
        model, alias, convert_fct=local_convert_fct,
        shape_fct=local_shape_fct, overwrite=overwrite,
        parser=parser, options=options)


def _internal_decorator(fct, op_version=None, runtime=None, signature=None,
                        register_class=None, overwrite=True, options=None):
    name = "onnxsklearn_parser_%s_%s_%s" % (
        fct.__name__, str(op_version), runtime)
    newclass = type(
        name, (wrapper_onnxnumpy_np,), {
            '__doc__': fct.__doc__,
            '__name__': name,
            '__getstate__': wrapper_onnxnumpy_np.__getstate__,
            '__setstate__': wrapper_onnxnumpy_np.__setstate__})
    _created_classes_inst.append(name, newclass)
    res = newclass(
        fct=fct, op_version=op_version, runtime=runtime,
        signature=signature)
    if register_class is not None:
        update_registered_converter_npy(
            register_class, "Sklearn%s" % getattr(
                register_class, "__name__", "noname"),
            res, shape_fct=None, overwrite=overwrite, options=options)
    return res


def onnxsklearn_transformer(op_version=None, runtime=None, signature=None,
                            register_class=None, overwrite=True):
    """
    Decorator to declare a converter for a transformer implemented using
    :epkg:`numpy` syntax but executed with :epkg:`ONNX`
    operators.

    :param op_version: :epkg:`ONNX` opset version
    :param runtime: `'onnxruntime'` or one implemented by @see cl OnnxInference
    :param signature: if None, the signature is replaced by a standard signature
        for transformer ``NDArraySameType("all")``
    :param register_class: automatically register this converter
        for this class to :epkg:`sklearn-onnx`
    :param overwrite: overwrite existing registered function if any

    .. versionadded:: 0.6
    """
    if signature is None:
        signature = NDArraySameType("all")

    def decorator_fct(fct):
        return _internal_decorator(fct, signature=signature,
                                   op_version=op_version,
                                   runtime=runtime,
                                   register_class=register_class,
                                   overwrite=overwrite)
    return decorator_fct


def onnxsklearn_regressor(op_version=None, runtime=None, signature=None,
                          register_class=None, overwrite=True):
    """
    Decorator to declare a converter for a regressor implemented using
    :epkg:`numpy` syntax but executed with :epkg:`ONNX`
    operators.

    :param op_version: :epkg:`ONNX` opset version
    :param runtime: `'onnxruntime'` or one implemented by @see cl OnnxInference
    :param signature: if None, the signature is replaced by a standard signature
        for transformer ``NDArraySameType("all")``
    :param register_class: automatically register this converter
        for this class to :epkg:`sklearn-onnx`
    :param overwrite: overwrite existing registered function if any

    .. versionadded:: 0.6
    """
    if signature is None:
        signature = NDArraySameType("all")

    def decorator_fct(fct):
        return _internal_decorator(fct, signature=signature,
                                   op_version=op_version,
                                   runtime=runtime,
                                   register_class=register_class,
                                   overwrite=overwrite)
    return decorator_fct


def onnxsklearn_classifier(op_version=None, runtime=None, signature=None,
                           register_class=None, overwrite=True):
    """
    Decorator to declare a converter for a classifier implemented using
    :epkg:`numpy` syntax but executed with :epkg:`ONNX`
    operators.

    :param op_version: :epkg:`ONNX` opset version
    :param runtime: `'onnxruntime'` or one implemented by @see cl OnnxInference
    :param signature: if None, the signature is replaced by a standard signature
        for transformer ``NDArraySameType("all")``
    :param register_class: automatically register this converter
        for this class to :epkg:`sklearn-onnx`
    :param overwrite: overwrite existing registered function if any

    .. versionadded:: 0.6
    """
    if signature is None:
        signature = NDArrayType(("T:all", ), dtypes_out=((numpy.int64, ), 'T'))

    def decorator_fct(fct):
        return _internal_decorator(fct, signature=signature,
                                   op_version=op_version,
                                   runtime=runtime,
                                   register_class=register_class,
                                   overwrite=overwrite,
                                   options={'zipmap': [False, True, 'columns'],
                                            'nocl': [False, True]})
    return decorator_fct


def onnxsklearn_cluster(op_version=None, runtime=None, signature=None,
                        register_class=None, overwrite=True):
    """
    Decorator to declare a converter for a cluster implemented using
    :epkg:`numpy` syntax but executed with :epkg:`ONNX`
    operators.

    :param op_version: :epkg:`ONNX` opset version
    :param runtime: `'onnxruntime'` or one implemented by @see cl OnnxInference
    :param signature: if None, the signature is replaced by a standard signature
        for transformer ``NDArraySameType("all")``
    :param register_class: automatically register this converter
        for this class to :epkg:`sklearn-onnx`
    :param overwrite: overwrite existing registered function if any

    .. versionadded:: 0.6
    """
    if signature is None:
        signature = NDArrayType(("T:all", ), dtypes_out=((numpy.int64, ), 'T'))

    def decorator_fct(fct):
        return _internal_decorator(fct, signature=signature,
                                   op_version=op_version,
                                   runtime=runtime,
                                   register_class=register_class,
                                   overwrite=overwrite)
    return decorator_fct


def _call_validate(self, X):
    if hasattr(self, "_validate_onnx_data"):
        return self._validate_onnx_data(X)
    return X


def _internal_method_decorator(register_class, method, op_version=None,
                               runtime=None, signature=None,
                               method_names=None, overwrite=True,
                               options=None):
    if isinstance(method_names, str):
        method_names = (method_names, )

    if issubclass(register_class, TransformerMixin):
        if signature is None:
            signature = NDArraySameType("all")
        if method_names is None:
            method_names = ("transform", )
    elif issubclass(register_class, RegressorMixin):
        if signature is None:
            signature = NDArraySameType("all")
        if method_names is None:
            method_names = ("predict", )
    elif issubclass(register_class, ClassifierMixin):
        if signature is None:
            signature = NDArrayType(
                ("T:all", ), dtypes_out=((numpy.int64, ), 'T'))
        if method_names is None:
            method_names = ("predict", "predict_proba")
        if options is None:
            options = {'zipmap': [False, True, 'columns'],
                       'nocl': [False, True]}
    elif issubclass(register_class, ClusterMixin):
        if signature is None:
            signature = NDArrayType(
                ("T:all", ), dtypes_out=((numpy.int64, ), 'T'))
        if method_names is None:
            method_names = ("predict", "transform")
    elif method_names is None:  # pragma: no cover
        raise RuntimeError(
            "No obvious API was detected (one among %s), "
            "then 'method_names' must be specified and not left "
            "empty." % (", ".join(map(lambda s: s.__name__, _default_cvt))))

    if method_names is None:
        raise RuntimeError(  # pragma: no cover
            "Methods to overwrite are not known for class %r and "
            "method %r." % (register_class, method))
    if signature is None:
        raise RuntimeError(  # pragma: no cover
            "Methods to overwrite are not known for class %r and "
            "method %r." % (register_class, method))

    name = "onnxsklearn_parser_%s_%s_%s" % (
        register_class.__name__, str(op_version), runtime)
    newclass = type(
        name, (wrapper_onnxnumpy_np,), {
            '__doc__': method.__doc__,
            '__name__': name,
            '__getstate__': wrapper_onnxnumpy_np.__getstate__,
            '__setstate__': wrapper_onnxnumpy_np.__setstate__})
    _created_classes_inst.append(name, newclass)

    def _check_(op):
        if isinstance(op, str):
            raise TypeError(  # pragma: no cover
                "Unexpected type: %r: %r." % (type(op), op))
        return op

    res = newclass(
        fct=lambda *args, op_=None, **kwargs: method(
            _check_(op_), *args, **kwargs),
        op_version=op_version, runtime=runtime, signature=signature,
        fctsig=method)

    if len(method_names) == 1:
        name = method_names[0]
        if hasattr(register_class, name):
            raise RuntimeError(  # pragma: no cover
                "Cannot overwrite method %r because it already exists in "
                "class %r." % (name, register_class))
        m = lambda self, X: res(_call_validate(self, X), op_=self)
        setattr(register_class, name, m)
    elif len(method_names) == 0:
        raise RuntimeError("No available method.")  # pragma: no cover
    else:
        m = lambda self, X: res(_call_validate(self, X), op_=self)
        setattr(register_class, method.__name__ + "_", m)
        for iname, name in enumerate(method_names):
            if hasattr(register_class, name):
                raise RuntimeError(  # pragma: no cover
                    "Cannot overwrite method %r because it already exists in "
                    "class %r." % (name, register_class))
            m = (lambda self, X, index_output=iname:
                 res(_call_validate(self, X), op_=self)[index_output])
            setattr(register_class, name, m)

    update_registered_converter_npy(
        register_class, "Sklearn%s" % getattr(
            register_class, "__name__", "noname"),
        res, shape_fct=None, overwrite=overwrite,
        options=options)
    return res


def onnxsklearn_class(method_name, op_version=None, runtime=None,
                      signature=None, method_names=None,
                      overwrite=True):
    """
    Decorator to declare a converter for a class derivated from
    :epkg:`scikit-learn`, implementing inference method
    and using :epkg:`numpy` syntax but executed with
    :epkg:`ONNX` operators.

    :param method_name: name of the method implementing the
        inference method with :epkg:`numpy` API for ONNX
    :param op_version: :epkg:`ONNX` opset version
    :param runtime: `'onnxruntime'` or one implemented by @see cl OnnxInference
    :param signature: if None, the signature is replaced by a standard signature
        depending on the model kind, otherwise, it is the signature of the
        ONNX function
    :param method_names: if None, method names is guessed based on
        the class kind (transformer, regressor, classifier, clusterer)
    :param overwrite: overwrite existing registered function if any

    .. versionadded:: 0.6
    """
    def decorator_class(objclass):
        _internal_method_decorator(
            objclass, method=getattr(objclass, method_name),
            signature=signature, op_version=op_version,
            runtime=runtime, method_names=method_names,
            overwrite=overwrite)
        return objclass

    return decorator_class



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunBinaryNumpy


class Min(OpRunBinaryNumpy):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunBinaryNumpy.__init__(self, numpy.minimum, onnx_node,
                                  desc=desc, **options)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ..shape_object import ShapeObject
from ._op import OpRun


def _global_average_pool(x):
    spatial_shape = numpy.ndim(x) - 2
    y = numpy.average(
        x, axis=tuple(range(spatial_shape, spatial_shape + 2)))
    for _ in range(spatial_shape):
        y = numpy.expand_dims(y, -1)
    return y


class GlobalAveragePool(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _run(self, x):  # pylint: disable=W0221
        res = _global_average_pool(x)
        return (res, )

    def _infer_shapes(self, x):  # pylint: disable=W0221
        if x.shape is None:
            return (ShapeObject(None, dtype=x.dtype), )
        shape = x.shape[:2] + (1, ) * (len(x.shape) - 2)
        return (ShapeObject(shape, dtype=x.dtype), )

    def _infer_types(self, x):  # pylint: disable=W0221
        return (x, )

    def _infer_sizes(self, *args):  # pylint: disable=W0221
        res = self.run(*args)
        return (dict(temp=0), ) + res



"""
@file
@brief :epkg:`numpy` functions implemented with :epkg:`onnx`
and compiled with this python runtime.

.. versionadded:: 0.6
"""
import numpy
from .onnx_numpy_annotation import (
    NDArrayType,
    NDArrayTypeSameShape,
    NDArraySameType,
    NDArraySameTypeSameShape)
from .numpy_onnx_impl import (
    abs as nx_abs,
    acos as nx_acos,
    acosh as nx_acosh,
    amin as nx_min,
    amax as nx_max,
    arange as nx_arange,
    argmax as nx_argmax,
    argmin as nx_argmin,
    asin as nx_asin,
    asinh as nx_asinh,
    atan as nx_atan,
    atanh as nx_atanh,
    ceil as nx_ceil,
    clip as nx_clip,
    compress as nx_compress,
    cos as nx_cos,
    cosh as nx_cosh,
    cumsum as nx_cumsum,
    concat as nx_concat,
    det as nx_det,
    dot as nx_dot,
    einsum as nx_einsum,
    erf as nx_erf,
    exp as nx_exp,
    expit as nx_expit,
    expand_dims as nx_expand_dims,
    floor as nx_floor,
    hstack as nx_hstack,
    isnan as nx_isnan,
    log as nx_log,
    mean as nx_mean,
    pad as nx_pad,
    prod as nx_prod,
    reciprocal as nx_reciprocal,
    relu as nx_relu,
    round as nx_round,
    sigmoid as nx_sigmoid,
    sign as nx_sign,
    sin as nx_sin,
    sinh as nx_sinh,
    sqrt as nx_sqrt,
    squeeze as nx_squeeze,
    sum as nx_sum,
    tan as nx_tan,
    tanh as nx_tanh,
    topk as nx_topk,
    transpose as nx_transpose,
    unsqueeze as nx_unsqueeze,
    vstack as nx_vstack,
    where as nx_where,
)
from .onnx_numpy_wrapper import onnxnumpy_np


@onnxnumpy_np(signature=NDArraySameTypeSameShape("all"))
def abs(x):
    "abs"
    return nx_abs(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def acos(x):
    "acos"
    return nx_acos(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def acosh(x):
    "acosh"
    return nx_acosh(x)


@onnxnumpy_np(signature=NDArrayType((numpy.int64, numpy.int64)))
def arange(start, stop, step=1):
    "arange, *start*, *stop* must be specified."
    return nx_arange(start, stop, step=step)


@onnxnumpy_np(signature=NDArraySameType("all"))
def amax(x, axis=None, keepdims=0):
    "amax"
    return nx_max(x, axis=axis, keepdims=keepdims)


@onnxnumpy_np(signature=NDArraySameType("all"))
def amin(x, axis=None, keepdims=0):
    "amin"
    return nx_min(x, axis=axis, keepdims=keepdims)


@onnxnumpy_np(signature=NDArrayType("all_int"))
def argmax(x, axis=0, keepdims=0):
    "argmax"
    return nx_argmax(x, axis=axis, keepdims=keepdims)


@onnxnumpy_np(signature=NDArrayType("all_int"))
def argmin(x, axis=0, keepdims=0):
    "argmin"
    return nx_argmin(x, axis=axis, keepdims=keepdims)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def asin(x):
    "asin"
    return nx_asin(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def asinh(x):
    "asinh"
    return nx_asinh(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def atan(x):
    "atan"
    return nx_atan(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def atanh(x):
    "atanh"
    return nx_atanh(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def ceil(x):
    "ceil"
    return nx_ceil(x)


@onnxnumpy_np(
    signature=NDArrayType(("all", "all", "all"), n_optional=2))
def clip(x, a_min=None, a_max=None):
    "clip"
    return nx_clip(x, a_min, a_max)


@onnxnumpy_np(signature=NDArrayType(("bool", "T:all"), dtypes_out=('T',)))
def compress(condition, x, axis=None):
    "compress"
    return nx_compress(condition, x, axis=axis)


@onnxnumpy_np(signature=NDArrayType("all", nvars=True))
def concat(*x, axis=0):
    "concat"
    return nx_concat(*x, axis=axis)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def cos(x):
    "cos"
    return nx_cos(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def cosh(x):
    "cosh"
    return nx_cosh(x)


@onnxnumpy_np(signature=NDArrayType(("all", "ints")))
def cumsum(x, axis):
    "cumsum"
    return nx_cumsum(x, axis)


@onnxnumpy_np(signature=NDArrayType("all"))
def det(x):
    "det"
    return nx_det(x)


@onnxnumpy_np(signature=NDArrayType(("T:all", "T")))
def dot(a, b):
    "dot"
    return nx_dot(a, b)


@onnxnumpy_np(signature=NDArrayType("all", nvars=True))
def einsum(*x, equation=None):
    "einsum"
    return nx_einsum(*x, equation=equation)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def erf(x):
    "erf"
    return nx_erf(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def exp(x):
    "exp"
    return nx_exp(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def expit(x):
    "expit"
    return nx_expit(x)


@onnxnumpy_np(signature=NDArrayType("floats"))
def expand_dims(x, axis=0):
    "expand_dims"
    return nx_expand_dims(x, axis)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def floor(x):
    "floor"
    return nx_floor(x)


@onnxnumpy_np(signature=NDArrayType("all", nvars=True))
def hstack(*x):
    "hstack"
    return nx_hstack(*x)


@onnxnumpy_np(signature=NDArrayTypeSameShape("all_bool"))
def isnan(x):
    "isnan"
    return nx_isnan(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def log(x):
    "log"
    return nx_log(x)


@onnxnumpy_np(signature=NDArrayType(("T:all", numpy.int64, 'T'), n_optional=1))
def pad(x, pads, constant_value=None, mode='constant'):
    "pad"
    return nx_pad(x, pads, mode=mode, constant_value=constant_value)


@onnxnumpy_np(signature=NDArraySameType("all"))
def prod(x, axis=None, keepdims=0):
    "prod"
    return nx_prod(x, axis=axis, keepdims=keepdims)


@onnxnumpy_np(signature=NDArraySameType("all"))
def mean(x, axis=None, keepdims=0):
    "mean"
    return nx_mean(x, axis=axis, keepdims=keepdims)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def reciprocal(x):
    "reciprocal"
    return nx_reciprocal(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def relu(x):
    "relu"
    return nx_relu(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def round(x):
    "round"
    return nx_round(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def sigmoid(x):
    "expit"
    return nx_sigmoid(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def sign(x):
    "sign"
    return nx_sign(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def sin(x):
    "sin"
    return nx_sin(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def sinh(x):
    "sinh"
    return nx_sinh(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def sqrt(x):
    "sqrt"
    return nx_sqrt(x)


@onnxnumpy_np(signature=NDArrayType(("all", numpy.int64), n_optional=1))
def squeeze(x, axis=None):
    "squeeze"
    return nx_squeeze(x, axis)


@onnxnumpy_np(signature=NDArraySameType("all"))
def sum(x, axis=None, keepdims=0):
    "sum"
    return nx_sum(x, axis=axis, keepdims=keepdims)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def tan(x):
    "tan"
    return nx_tan(x)


@onnxnumpy_np(signature=NDArraySameTypeSameShape("floats"))
def tanh(x):
    "tanh"
    return nx_tanh(x)


@onnxnumpy_np(signature=NDArrayType(("T:all", "ints"), ("T", (numpy.int64,))))
def topk(x, k, axis=-1, largest=1, sorted=1):
    "topk"
    return nx_topk(x, k, axis=axis, largest=largest, sorted=sorted)


@onnxnumpy_np(signature=NDArraySameType("all"))
def transpose(x, perm=(1, 0)):
    "transpose"
    return nx_transpose(x, perm=perm)


@onnxnumpy_np(signature=NDArrayType(("all", numpy.int64)))
def unsqueeze(x, axes):
    "unsqueeze"
    return nx_unsqueeze(x, axes)


@onnxnumpy_np(signature=NDArrayType("all", nvars=True))
def vstack(*x):
    "vstack"
    return nx_vstack(*x)


@onnxnumpy_np(signature=NDArrayType(("bool", "T:all", "T"), ("T", )))
def where(cond, x, y):
    "where"
    return nx_where(cond, x, y)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.defs import onnx_opset_version
from ._op import OpRun


def _batchnorm_test_mode(x, s, bias, mean, var, epsilon=1e-5):
    dims_x = len(x.shape)
    dim_ones = (1,) * (dims_x - 2)
    s = s.reshape(-1, *dim_ones)
    bias = bias.reshape(-1, *dim_ones)
    mean = mean.reshape(-1, *dim_ones)
    var = var.reshape(-1, *dim_ones)
    y = s * (x - mean) / numpy.sqrt(var + epsilon) + bias
    return y.astype(x.dtype)


def _batchnorm_training_mode(x, s, bias, mean, var, momentum=0.9,
                             epsilon=1e-5):
    axis = tuple(numpy.delete(numpy.arange(len(x.shape)), 1))
    saved_mean = x.mean(axis=axis)
    saved_var = x.var(axis=axis)
    output_mean = mean * momentum + saved_mean * (1 - momentum)
    output_var = var * momentum + saved_var * (1 - momentum)
    y = _batchnorm_test_mode(x, s, bias, saved_mean, saved_var,
                             epsilon=epsilon)
    return (y.astype(x.dtype), saved_mean.astype(x.dtype),
            saved_var.astype(x.dtype), output_mean.astype(x.dtype),
            output_var.astype(x.dtype))


class BatchNormalization_9(OpRun):

    atts = {'epsilon': 1e-5, 'momentum': 0.9}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=BatchNormalization.atts,
                       **options)

    def _run(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        res = _batchnorm_test_mode(
            x, scale, bias, mean, var, epsilon=self.epsilon)
        return (res, )

    def _infer_shapes(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        return (x, )

    def _infer_types(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        return (x, )

    def _infer_sizes(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        res = self.run(x, scale, bias, mean, var)
        return (dict(temp=x.size * x.dtype.itemsize * 2), ) + res


class BatchNormalization_14(OpRun):

    atts = {'epsilon': 1e-5, 'momentum': 0.9, 'training_mode': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=BatchNormalization.atts,
                       **options)

    def _run(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        if self.training_mode == 0:
            res = _batchnorm_test_mode(
                x, scale, bias, mean, var, epsilon=self.epsilon)
            return (res, )
        res, saved_mean, saved_var, output_mean, output_var = (
            _batchnorm_training_mode(x, scale, bias, mean, var,
                                     self.momentum, self.epsilon))
        return res, saved_mean, saved_var, output_mean, output_var

    def _infer_shapes(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        if self.training_mode == 0:
            return (x, )
        return (x, scale, bias, mean, var)

    def _infer_types(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        if self.training_mode == 0:
            return (x, )
        return (x, scale, bias, mean, var)

    def _infer_sizes(self, x, scale, bias, mean, var):  # pylint: disable=W0221
        if self.training_mode == 0:
            res = self.run(x, scale, bias, mean, var)
            return (dict(temp=x.size * x.dtype.itemsize * 2), ) + res
        res = self.run(x, scale, bias, mean, var)
        return (dict(temp=x.size * x.dtype.itemsize * 4), ) + res


if onnx_opset_version() >= 14:
    BatchNormalization = BatchNormalization_14
else:  # pragma: no cover
    BatchNormalization = BatchNormalization_9



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ..shape_object import ShapeObject


class ConcatFromSequence(OpRun):

    atts = {'axis': 0, 'new_axis': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=ConcatFromSequence.atts,
                       **options)

    def _run(self, seq):  # pylint: disable=W0221
        if self.new_axis == 1:
            seq2 = [s[..., numpy.newaxis] for s in seq]
            res = numpy.concatenate(seq2, axis=-1)
        else:
            res = numpy.concatenate(seq, axis=self.axis)
        return (res, )

    def _infer_shapes(self, seq):  # pylint: disable=W0221
        return (ShapeObject(None, seq.dtype), )

    def _infer_types(self, seq):  # pylint: disable=W0221
        return (seq, )

    def _infer_sizes(self, seq):  # pylint: disable=W0221
        res = self.run(seq)
        if self.new_axis == 1:
            return (dict(temp=sum(o.size for o in seq)), ) + res
        return (dict(temp=0), ) + res



"""
@file
@brief Common functions to reduce the number of
nodes of an :epkg:`ONNX` graphs.
"""
from onnx.helper import make_graph, ValueInfoProto, make_model
from onnx import AttributeProto, NodeProto
from onnx.helper import make_attribute


def _apply_optimisation_on_graph(fct, onnx_model, recursive=True, debug_info=None,
                                 **kwargs):
    """
    Applies an optimisation function *fct* on a graph
    and not on the model.

    @param      fct             function to optimise like
                                @see fn onnx_remove_node_identity
    @param      onnx_model      onnx model
    @param      recursive       looks into subgraphs
    @param      debug_info      debug information (private)
    @param      kwargs          additional parameters
    @return                     new onnx _model
    """
    if hasattr(onnx_model, 'graph'):
        if debug_info is None:
            debug_info = []
        graph = fct(
            onnx_model.graph, debug_info=debug_info + ['GRAPH'],
            **kwargs)
        new_model = make_model(graph)
        new_model.ir_version = onnx_model.ir_version
        new_model.producer_name = onnx_model.producer_name
        new_model.producer_version = onnx_model.producer_version
        new_model.domain = onnx_model.domain
        new_model.model_version = onnx_model.model_version
        new_model.doc_string = onnx_model.doc_string
        if hasattr(onnx_model, 'value_info'):
            graph.value_info.extend(onnx_model.value_info)  # pragma: no cover
        while len(new_model.opset_import) > 0:  # pylint: disable=E1101
            new_model.opset_import.pop()  # pylint: disable=E1101
        for oimp in onnx_model.opset_import:
            op_set = new_model.opset_import.add()  # pylint: disable=E1101
            op_set.domain = oimp.domain
            op_set.version = oimp.version
        return new_model
    raise TypeError(  # pragma: no cover
        "This function only works on 'ModelProto' anod not not on"
        " {}.".format(type(onnx_model)))


def _apply_remove_node_fct_node(fct, node, recursive, debug_info):
    """
    Applies an optimizing function on a subgraphs.

    @param      node        onnx node
    @param      recursive   does it in subgraphs as well
    @return                 new node
    """
    if not hasattr(node, 'attribute'):
        return node  # pragma: no cover
    modified = 0
    new_atts = []
    for att in node.attribute:
        if att.name == 'body':
            new_body = fct(
                att.g, recursive=recursive,
                debug_info=debug_info + [att.name])
            new_atts.append(_make_att_graph(att.name, new_body))
            modified += 1
        else:
            new_atts.append(att)
    if modified > 0:
        new_node = _make_node(node.op_type, node.input,
                              node.output, name=node.name,
                              attributes=new_atts)
        return new_node
    return node


def _make_node(op_type, inputs, outputs, name=None, doc_string=None,
               domain=None, attributes=None):
    """
    Constructs a NodeProto.

    :param op_type: (string): The name of the operator to construct
    :param inputs: list of input names
    :param outputs: list of output names
    :param name: optional unique identifier for NodeProto
    :param doc_string: optional documentation
        string for NodeProto
    :param domain: optional domain for NodeProto.
        If it's None, we will just use default domain (which is empty)
    :param attributes: the attributes of the node. The acceptable values
        are documented in :func:`make_attribute`.
    :return: node
    """
    node = NodeProto()
    node.op_type = op_type
    node.input.extend(inputs)  # pylint: disable=E1101
    node.output.extend(outputs)  # pylint: disable=E1101
    if name:
        node.name = name
    if doc_string:
        node.doc_string = doc_string  # pragma: no cover
    if domain is not None:
        node.domain = domain
    if isinstance(attributes, dict):
        if len(attributes) > 0:  # pragma: no cover
            node.attribute.extend(  # pylint: disable=E1101
                make_attribute(key, value)
                for key, value in sorted(attributes.items()))
    elif attributes:
        for att in attributes:
            node.attribute.extend([att])  # pylint: disable=E1101
    return node


def _replace(name, old_name, new_name):
    if isinstance(old_name, dict) and new_name is None:
        return old_name.get(name, name)
    if name == old_name:
        return new_name
    return name


def _rename_node_input(onnx_node, old_name, new_name=None):
    """
    Renames an input from a node.

    @param      onnx_node       onnx_node
    @param      old_name        old name
    @param      new_name        new name or None if *old_name* is a dictionary
    @return                     new node
    """
    inputs = [_replace(name, old_name, new_name) for name in onnx_node.input]
    outputs = list(onnx_node.output)
    if hasattr(onnx_node, 'attribute'):
        new_atts = []
        for att in onnx_node.attribute:
            if att.name == 'body':
                new_body = _rename_graph_input(att.g, old_name, new_name)
                attr = AttributeProto()
                attr.name = att.name
                attr.g.CopyFrom(new_body)  # pylint: disable=E1101
                attr.type = AttributeProto.GRAPH  # pylint: disable=E1101
                new_atts.append(attr)
            else:
                new_atts.append(att)
        atts = new_atts
    else:
        atts = None  # pragma: no cover
    node = _make_node(
        onnx_node.op_type, inputs, outputs, name=onnx_node.name,
        domain=onnx_node.domain, attributes=atts)
    return node


def _copy_value_info_proto(new_name, obj):
    value_info = ValueInfoProto()
    value_info.name = new_name
    value_info.type.CopyFrom(obj.type)  # pylint: disable=E1101
    if obj.type.doc_string:
        value_info.doc_string = obj.type.doc_string
    return value_info


def _rename_graph_output(graph, old_name, new_name):
    """
    Renames an output and adds an *Identity* node
    to connect the dots.

    @param      graph       ONNX graph
    @return                 modified graph
    """
    outputs = []
    for o in graph.output:
        if old_name != o.name:
            outputs.append(o)
        else:
            outputs.append(_copy_value_info_proto(new_name, o))
    nodes = list(graph.node)
    nodes.append(_make_node('Identity', [old_name], [new_name]))
    new_graph = make_graph(nodes, graph.name, graph.input, outputs,
                           graph.initializer)
    new_graph.value_info.extend(graph.value_info)  # pylint: disable=E1101
    return new_graph


def _rename_graph_input(graph, old_name, new_name):
    """
    Renames an input and adds an *Identity* node
    to connect the dots.

    @param      graph       ONNX graph
    @return                 modified graph
    """
    inputs = []
    for i in graph.input:
        if old_name != i.name:
            inputs.append(i)
        else:
            inputs.append(_copy_value_info_proto(new_name, i))
    nodes = list(graph.node)
    nodes.append(_make_node('Identity', [new_name], [old_name]))
    new_graph = make_graph(nodes, graph.name, inputs, graph.output,
                           graph.initializer)
    new_graph.value_info.extend(graph.value_info)  # pylint: disable=E1101
    return new_graph


def _make_att_graph(name, new_body):
    attr = AttributeProto()
    attr.name = name
    attr.g.CopyFrom(new_body)  # pylint: disable=E1101
    attr.type = AttributeProto.GRAPH  # pylint: disable=E1101
    return attr


def _rename_node_output(onnx_node, old_name, new_name):
    """
    Renames an output from a node.

    @param      onnx_node       onnx_node
    @param      old_name        old name
    @param      new_name        new name
    @return                     new node
    """
    inputs = list(onnx_node.input)
    outputs = [_replace(name, old_name, new_name) for name in onnx_node.output]
    if hasattr(onnx_node, 'attribute'):
        new_atts = []
        for att in onnx_node.attribute:
            if att.name == 'body':
                new_body = _rename_graph_output(att.g, old_name, new_name)
                new_atts.append(_make_att_graph(att.name, new_body))
            else:
                new_atts.append(att)
        atts = new_atts
    else:
        atts = None  # pragma: no cover
    node = _make_node(
        onnx_node.op_type, inputs, outputs, name=onnx_node.name,
        domain=onnx_node.domain, attributes=atts)
    return node



"""
@file
@brief Identifies a version of a function.

.. versionadded:: 0.6
"""
from collections import namedtuple


class FctVersion(namedtuple("_version_", ['args', 'kwargs'])):
    """
    Identifies a version of a function based on its
    arguments and its parameters.
    """
    __slots__ = ()

    def _check_(self):
        if self.args is not None and not isinstance(self.args, tuple):
            raise TypeError("args must be None or a tuple.")
        if self.kwargs is not None and not isinstance(self.kwargs, tuple):
            raise TypeError("kwargs must None or be a tuple.")

    def __repr__(self):
        "usual"
        def cl(s):
            return str(s).replace("<class '", "").replace("'>", "")
        if self.args is None:
            sa = "None"
        else:
            sa = ",".join(map(cl, self.args))
            sa = ("(%s)" % sa) if len(self.args) > 1 else ("(%s,)" % sa)

        return "%s(%s, %s)" % (
            self.__class__.__name__, sa, self.kwargs)

    def __len__(self):
        "Returns the sum of lengths."
        return ((0 if self.args is None else len(self.args)) +
                (0 if self.kwargs is None else len(self.kwargs)))

    def as_tuple(self):
        "Returns a single tuple for the version."
        return ((tuple() if self.args is None else self.args) +
                (tuple() if self.kwargs is None else self.kwargs))

    def as_tuple_with_sep(self, sep):
        "Returns a single tuple for the version."
        return ((tuple() if self.args is None else self.args) +
                (sep, ) +
                (tuple() if self.kwargs is None else self.kwargs))

    def as_string(self):
        "Returns a single stirng identifier."
        val = "_".join(map(str, self.as_tuple_with_sep("_")))
        val = val.replace("<class 'numpy.", "").replace(
            '.', "_").replace("'>", "").replace(" ", "")
        return val.lower()



"""
@file
@brief Documentation helper.
"""
from logging import getLogger
from textwrap import indent, dedent
import numpy
from jinja2 import Template
from pandas import DataFrame, notnull
from sklearn.linear_model import LinearRegression
from pyquickhelper.loghelper import noLOG
from pyquickhelper.pandashelper.tblformat import df2rst
from sklearn import __all__ as sklearn__all__
from ...tools.asv_options_helper import get_opset_number_from_onnx
from ...tools.model_info import analyze_model
from ..validate.validate import enumerate_validated_operator_opsets, sklearn_operators
from ...onnx_tools.optim.sklearn_helper import inspect_sklearn_model
from ...onnx_tools.optim.onnx_helper import onnx_statistics
from ..onnx_inference import OnnxInference
from ..validate.validate_summary import _clean_values_optim
from .doc_helper import visual_rst_template


def _make_opset(row):
    opsets = []
    if hasattr(row, 'to_dict'):
        row = row.to_dict()  # pragma: no cover
    for k, v in row.items():
        if k.startswith('opset'):
            if isinstance(v, int):
                opsets.append(v)
            elif isinstance(v, float):  # pragma: no cover
                if numpy.isnan(v):
                    opsets.append(0)
                else:
                    opsets.append(int(v))
            else:  # pragma: no cover
                vv = list(_ for _ in v if 'OK' in str(v))
                if len(vv) > 0:
                    opsets.append(int(k.replace("opset", "")))
    if len(opsets) == 0:
        return "o%d" % get_opset_number_from_onnx()  # pragma: no cover
    val = max(opsets)
    return "o%d" % val


def enumerate_visual_onnx_representation_into_rst(sub, fLOG=noLOG):
    """
    Returns content for pages such as
    :ref:`l-skl2onnx-linear_model`.
    """
    logger = getLogger('skl2onnx')
    logger.disabled = True

    rst_templ = visual_rst_template()
    templ = Template(rst_templ)
    done = set()
    subsets = [_['name'] for _ in sklearn_operators(sub)]
    subsets.sort()
    for row in enumerate_validated_operator_opsets(
            verbose=0, debug=None, fLOG=fLOG,
            opset_min=get_opset_number_from_onnx(),
            opset_max=get_opset_number_from_onnx(),
            store_models=True, models=subsets):

        if 'ONNX' not in row:
            continue
        name = row['name']
        scenario = row['scenario']
        problem = row['problem']
        model = row['MODEL']
        method = row['method_name']
        optim = row.get('optim', '')
        opset = _make_opset(row)
        stats_skl = inspect_sklearn_model(model)
        stats_onx = onnx_statistics(row['ONNX'])
        stats_model = analyze_model(model)
        stats = {'skl_' + k: v for k, v in stats_skl.items()}
        stats.update({'onx_' + k: v for k, v in stats_onx.items()})
        stats.update({'fit_' + k: v for k, v in stats_model.items()})

        df = DataFrame([stats])
        table = df2rst(df.T.reset_index(drop=False))

        clean_optim = _clean_values_optim(optim)
        title = " - ".join([name, problem, scenario, clean_optim])
        if title in done:
            continue  # pragma: no cover
        done.add(title)
        link = "-".join([name, problem, scenario, clean_optim, opset])
        link = link.replace(" ", "").replace(
            "{", "").replace("}", "").replace("'", "")

        optim_param = ("Model was converted with additional parameter: ``{}``.".format(optim)
                       if optim else "")

        oinf = OnnxInference(row['ONNX'], skip_run=True)
        dot = oinf.to_dot(recursive=True)
        try:
            res = templ.render(dot=dot, model=repr(model), method=method,
                               kind=problem, title=title,
                               indent=indent, len=len,
                               link=link, table=table,
                               optim_param=optim_param)
        except KeyError as e:  # pragma: no cover
            rows = [
                '', str(e), '',
                "title='{}'".format(title),
                "method='{}'".format(method),
                "problem='{}'".format(problem),
                model.__class__.__name__, "", "---------",
                rst_templ]
            res = ".. index:: docissue:\n\n::\n\n" + \
                indent("\n".join(rows), "    ")
        yield res


def compose_page_onnxrt_ops(level="^"):
    """
    Writes page :ref:`l-onnx-runtime-operators`.

    @param      level       title level
    """
    begin = dedent("""
    .. _l-onnx-runtime-operators:

    Python Runtime for ONNX operators
    =================================

    The main function instantiates a runtime class which
    computes the outputs of a specific node.

    .. autosignature:: mlprodict.onnxrt.ops.load_op

    Other sections documents available operators.
    This project was mostly started to show a way to
    implement a custom runtime, do some benchmarks,
    test, exepriment...

    .. contents::
        :local:

    Python
    ++++++

    """)
    from ..ops_cpu._op_list import _op_list

    names = []
    for op in _op_list:
        names.append((op.__name__, op))
    names.sort()

    rows = [begin]
    for name, op in names:
        rows.append("")
        rows.append(".. _lpyort-{}:".format(name))
        rows.append("")
        rows.append(name)
        rows.append(level * len(name))
        rows.append("")
        mod = op.__module__.split('.')[-1]
        rows.append(
            ".. autosignature:: mlprodict.onnxrt.ops_cpu.{}.{}".format(mod, name))
        rows.append('')
    return "\n".join(rows)


def split_columns_subsets(df):
    """
    Functions used in the documentation to split
    a dataframe by columns into multiple dataframe to
    reduce the scrolling.
    """
    common = [c for c in ['name', 'problem',
                          'scenario', 'optim'] if c in df.columns]
    subsets = []
    subsets.append(
        [c for c in df.columns if 'opset' in c or 'onx_nnodes' == c])
    subsets.append([c for c in df.columns if 'ERROR' in c or 'opset' in c])
    subsets.append([c for c in df.columns if c.startswith(
        'skl_') or c.startswith('onx_') or 'opset' in c])
    subsets.append([c for c in df.columns if 'N=' in c or 'opset' in c])
    subsets = [s for s in subsets if len(s) > 0]
    return common, subsets


def build_key_split(key, index):
    """
    Used for documentation.
    """
    try:
        new_key = str(key).split('`')[1].split('<')[0].strip()
    except IndexError:
        new_key = str(key)
    if 'SVC' in new_key or 'SVR' in new_key or 'SVM' in new_key:
        return 'SVM'  # pragma: no cover
    if 'Neighbors' in new_key:
        return 'Neighbors'  # pragma: no cover
    if 'Scaler' in new_key:
        return 'Scaler'  # pragma: no cover
    if 'Normalizer' in new_key:
        return 'Scaler'  # pragma: no cover
    if new_key.endswith("NB"):
        return "...NB"  # pragma: no cover
    if new_key.endswith("RBM"):
        return "...NB"  # pragma: no cover
    if "KMeans" in new_key:
        return "KMeans"  # pragma: no cover
    if ('XGB' in new_key or 'LGBM' in new_key or 'Tree' in new_key or
            'Forest' in new_key):
        return 'Trees'  # pragma: no cover
    if ('ARDRegression' in new_key or 'ElasticNet' in new_key or
            'HuberRegressor' in new_key or 'Lars' in new_key or
            'Lasso' in new_key or 'LinearRegression' in new_key or
            'LogisticRegression' in new_key or
            'Ridge' in new_key or 'SGD' in new_key or
            'TheilSen' in new_key):
        return 'Linear'  # pragma: no cover
    for begin in ["Lasso", "Select", "Label", 'Tfidf', 'Feature',
                  'Bernoulli', 'MultiTask', 'OneVs', 'PLS',
                  'Sparse', 'Spectral', 'MiniBatch',
                  'Bayesian']:
        if new_key.startswith(begin):
            return begin + '...'
    for end in ['CV', 'Regressor', 'Classifier']:  # pragma: no cover
        if new_key.endswith(end):
            new_key = new_key[:-len(end)]
    return new_key  # pragma: no cover


def filter_rows(df):
    """
    Used for documentation.
    """
    for c in ['ERROR-msg', 'RT/SKL-N=1']:
        if c in df.columns:
            return df[df[c].apply(lambda x: notnull(x) and x not in (None, '', 'nan'))]
    return df



"""
@file
@brief Loads runtime operator.
"""


def load_op(onnx_node, desc=None, options=None, variables=None, dtype=None):
    """
    Sets up a class for a specific ONNX operator.

    @param      onnx_node       :epkg:`onnx` node
    @param      desc            internal representation
    @param      options         runtime options
    @param      variables       registered variables created by previous operators
    @param      dtype           float computational type
    @return                     runtime class
    """
    if desc is None:
        raise ValueError(  # pragma: no cover
            "desc should not be None.")
    if options is None:
        provider = 'python'  # pragma: no cover
    else:
        provider = options.get('provider', 'python')
        if 'provider' in options:
            options = options.copy()
            del options['provider']
    if provider == 'python':
        from .ops_cpu import load_op as lo
        return lo(onnx_node, desc=desc, options=options)
    if provider == 'empty':
        from .ops_empty import load_op as lo
        return lo(onnx_node, desc=desc, options=options)
    if provider == 'onnxruntime2':
        from .ops_onnxruntime import load_op as lo
        return lo(onnx_node, desc=desc, options=options,  # pylint: disable=E1123
                  variables=variables, dtype=dtype)
    raise ValueError("Unable to handle provider '{}'.".format(provider))



"""
@file
@brief Function to measure the performance of einsum decomposition.
"""
from itertools import permutations
import numpy
from onnx import helper, TensorProto
from ...tools.ort_wrapper import InferenceSession
from ...onnxrt import OnnxInference
from ..bench_helper import measure_time
from .einsum_impl import decompose_einsum_equation, apply_einsum_sequence


def _make_einsum_model(equation, opset=14):  # opset=13, 14, ...
    from skl2onnx.common._topology import OPSET_TO_IR_VERSION  # pylint: disable=E0611,E0001
    inputs = equation.split('->')[0].split(',')

    model = helper.make_model(
        opset_imports=[helper.make_operatorsetid('', opset)],
        ir_version=OPSET_TO_IR_VERSION.get(opset, 7),
        producer_name='mlprodict',
        producer_version='0.1',
        graph=helper.make_graph(
            name='einsum_test',
            inputs=[
                helper.make_tensor_value_info(
                    "X%d" % i, TensorProto.FLOAT, None)  # pylint: disable=E1101
                for i in range(len(inputs))],
            outputs=[
                helper.make_tensor_value_info(
                    "Y", TensorProto.FLOAT, None)],  # pylint: disable=E1101
            nodes=[
                helper.make_node(
                    "Einsum", ["X%d" % i for i in range(len(inputs))], ["Y"],
                    equation=equation)
            ]
        )
    )
    return model


def _make_inputs(equation, shapes):
    inputs = equation.split('->')[0].split(',')
    dims = [len(i) for i in inputs]

    if isinstance(shapes, int):
        N = shapes
        shapes = [(N, ) * le for le in dims]
    else:
        if len(shapes) != len(inputs):
            raise ValueError(  # pragma: no cover
                "Unexpected number of shapes %r with equation %r."
                "" % (shapes, equation))
    inputs = [numpy.random.randn(*sh) for sh in shapes]
    return [i.astype(numpy.float32) for i in inputs]


def einsum_benchmark(equation="abc,cd->abd", shape=30, perm=False,
                     runtime='python', use_tqdm=False,
                     number=5, repeat=5, opset=14):  # opset=13, 14, ...
    """
    Investigates whether or not the decomposing einsum is faster.

    :param equation: einsum equation to test
    :param shape: an integer (all dimension gets the same size) or
        a list of shapes in a string separated with `;`)
    :param perm: check on permutation or all letter permutations
    :param runtime: numpy, python, onnxruntime
    :param use_tqdm: show progress
    :param output: output file (usually a csv file or an excel file),
        it requires pandas
    :param number: usual parameter to measure a function
    :param repeat: usual parameter to measure a function
    :param opset: target opset
    :return: list of dictionaries as an iterator
    """
    scenarios = []
    if (isinstance(shape, list) and
            all(map(lambda t: isinstance(t, int), shape))):
        shape_list = shape
    else:
        shape_list = [shape]

    if perm:
        if equation.lower() != equation:
            raise ValueError(
                "Only equations with lower letters are allowed but equation %r "
                "is not." % equation)
        letters = list(sorted(set(
            c for c in equation if "a" <= c < "z" or "A" <= c < "Z")))
        for p in permutations(letters):
            replace = {d: c for c, d in zip(letters, p)}
            eq = equation
            for k, v in replace.items():
                eq = eq.replace(k, v.upper())
            eq = eq.lower()
            for dec in ['einsum', 'dec']:
                for sh in shape_list:
                    scenarios.append((eq, runtime, dec, sh))
    else:
        for dec in ['einsum', 'dec']:
            for sh in shape_list:
                scenarios.append((equation, runtime, dec, sh))

    if use_tqdm:
        from tqdm import tqdm  # pragma: no cover
        loop = tqdm(scenarios)  # pragma: no cover
    else:
        loop = scenarios

    for eq, rt, dec, sh in loop:
        inputs = _make_inputs(equation, sh)

        if dec == 'dec':
            seq = decompose_einsum_equation(eq, strategy='numpy', clean=True)
        else:
            seq = None

        if rt == 'numpy':
            if dec == 'einsum':
                fct = lambda *x, eq=eq: numpy.einsum(eq, *x, optimize=True)
            else:
                fct = lambda *x, seq=seq: apply_einsum_sequence(seq, *x)
        elif rt == 'onnxruntime':
            if dec == 'einsum':
                onx = _make_einsum_model(equation, opset=opset)
            else:
                onx = seq.to_onnx('Y', *["X%d" % i for i in range(len(inputs))],
                                  opset=opset)
            sess = InferenceSession(
                onx.SerializeToString())  # pylint: disable=W0612
            fct = lambda *x, se=sess: se.run(
                None, {"X%d" % i: v for i, v in enumerate(x)})
        elif rt == 'python':
            if dec == 'einsum':
                onx = _make_einsum_model(equation, opset=opset)
            else:
                onx = seq.to_onnx('Y', *["X%d" % i for i in range(len(inputs))],
                                  opset=opset)
            oinf = OnnxInference(onx)  # pylint: disable=W0612
            fct = lambda *x, oi=oinf: oi.run(
                {"X%d" % i: v for i, v in enumerate(x)})
        else:
            raise ValueError("Unexpected runtime %r." % rt)

        res = measure_time(fct, *inputs, repeat=repeat, number=number)
        res['rt'] = rt
        res['dec'] = dec
        res['eq'] = eq
        res['shapes'] = ";".join(
            map(str, [m.shape for m in inputs])).replace(' ', '')
        yield res



# -*- coding: utf-8 -*-
"""
@file
@brief Implements command line ``python -m mlprodict <command> <args>``.
"""
import sys
import warnings


def main(args, fLOG=print):
    """
    Implements ``python -m mlprodict <command> <args>``.

    @param      args        command line arguments
    @param      fLOG        logging function
    """
    try:
        from .cli.validate import validate_runtime
        from .cli.convert_validate import convert_validate
        from .cli.optimize import onnx_optim, onnx_stats
        from .cli.asv_bench import asv_bench
        from .cli.asv2csv import asv2csv
        from .cli.replay import benchmark_replay
        from .cli.einsum import einsum_test
    except ImportError:  # pragma: no cover
        from mlprodict.cli.validate import validate_runtime
        from mlprodict.cli.convert_validate import convert_validate
        from mlprodict.cli.optimize import onnx_optim, onnx_stats
        from mlprodict.cli.asv_bench import asv_bench
        from mlprodict.cli.asv2csv import asv2csv
        from mlprodict.cli.replay import benchmark_replay
        from mlprodict.cli.einsum import einsum_test

    fcts = dict(validate_runtime=validate_runtime,
                convert_validate=convert_validate,
                onnx_optim=onnx_optim,
                onnx_stats=onnx_stats,
                asv_bench=asv_bench,
                asv2csv=asv2csv,
                benchmark_replay=benchmark_replay,
                einsum_test=einsum_test)
    try:
        from pyquickhelper.cli import cli_main_helper
    except ImportError:  # pragma: no cover
        warnings.warn("The command line requires module pyquickhelper.")
        return None
    return cli_main_helper(fcts, args=args, fLOG=fLOG)


if __name__ == "__main__":
    main(sys.argv[1:])  # pragma: no cover



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Loop(OpRun):

    atts = {
        'body': None,
    }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Loop.atts,
                       **options)
        if not hasattr(self.body, 'run'):
            raise RuntimeError("Parameter 'body' must have a method 'run', "
                               "type {}.".format(type(self.body)))

        self._run_meth = (self.body.run_in_scan
                          if hasattr(self.body, 'run_in_scan')
                          else self.body.run)

    def _run(self, M, cond, v_initial, *args):  # pylint: disable=W0221
        inputs = {name: None for name in self.body.input_names}
        inputs[self.body.input_names[2]] = v_initial
        cond_name = self.body.output_names[1]
        if len(args) > 0:
            begin = len(self.body.input_names) - len(args)
            for name, val in zip(self.body.input_names[begin:], args):
                inputs[name] = val
        it = 0
        while cond and it < M:
            inputs[self.body.input_names[0]] = numpy.array(it, dtype=M.dtype)
            inputs[self.body.input_names[1]] = cond
            outputs = self._run_meth(inputs)
            cond = outputs[cond_name]
            for i, o in zip(self.body.input_names[2:],
                            self.body.output_names[1:]):
                inputs[i] = outputs[o]
            it += 1
        if it == 0:
            outputs = {self.body.output_names[1]: cond}
            for i, o in zip(self.body.input_names[2:],
                            self.body.output_names[1:]):
                outputs[o] = inputs[i]
        for o in self.body.output_names:
            if o not in outputs:
                outputs[o] = numpy.empty(shape=tuple())
        return tuple([outputs[name] for name in self.body.output_names[1:]])

    def _infer_shapes(self, M, cond, v_initial, *args):  # pylint: disable=W0221
        res = self.body._set_shape_inference_runtime()
        return tuple([res[name] for name in self.body.output_names[1:]])

    def _infer_types(self, M, cond, v_initial, *args):  # pylint: disable=W0221
        res = self.body._set_type_inference_runtime()
        return tuple([res[name] for name in self.body.output_names[1:]])



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.defs import onnx_opset_version
from ._op import OpRunReduceNumpy, RuntimeTypeError, OpRun


class ReduceSum_1(OpRunReduceNumpy):

    atts = {'axes': [], 'keepdims': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunReduceNumpy.__init__(self, onnx_node, desc=desc,
                                  expected_attributes=ReduceSum_1.atts,
                                  **options)

    def _run(self, data):  # pylint: disable=W0221
        return (numpy.sum(data, axis=self.axes,
                          keepdims=self.keepdims,
                          dtype=data.dtype), )


class ReduceSum_11(ReduceSum_1):

    def __init__(self, onnx_node, desc=None, **options):
        ReduceSum_1.__init__(self, onnx_node, desc=desc, **options)


class ReduceSum_13(OpRunReduceNumpy):

    atts = {'axes': [], 'keepdims': 1, 'noop_with_empty_axes': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunReduceNumpy.__init__(self, onnx_node, desc=desc,
                                  expected_attributes=ReduceSum_13.atts,
                                  **options)

    def run(self, data, axes=None):  # pylint: disable=E0202,W0221
        """
        Calls method ``_run``.
        """
        res = self._run(data, axes=axes)
        if not self.keepdims and not isinstance(res[0], numpy.ndarray):
            res = (numpy.array([res[0]], dtype=res[0].dtype), )
        if res[0].dtype != data.dtype:
            raise RuntimeTypeError(  # pragma: no cover
                "Output type mismatch: input '{}' != output '{}' "
                "(operator '{}')".format(
                    data.dtype, res[0].dtype, self.__class__.__name__))
        return res

    def _run_no_checks_(self, x, axes=None):  # pylint: disable=W0221
        return OpRun.run(self, x, axes)

    def _run(self, data, axes=None):  # pylint: disable=W0221
        if axes is None and self.noop_with_empty_axes:
            return (data, )
        if axes is not None and not isinstance(axes, int):
            if isinstance(axes, numpy.ndarray) and len(axes.shape) == 0:
                axes = int(axes)
            else:
                axes = tuple(axes) if len(axes) > 0 else None
        try:
            return (numpy.sum(data, axis=axes,
                              keepdims=self.keepdims,
                              dtype=data.dtype), )
        except TypeError as e:  # pragma: no cover
            raise TypeError(
                "Unable to reduce shape %r with axes=%r." % (
                    data.shape, axes)) from e

    def infer_shapes(self, data, axes=None):  # pylint: disable=E0202,W0221
        return self._infer_shapes(data, axes=axes)

    def _infer_shapes(self, data, axes=None):  # pylint: disable=W0221
        """
        Returns the same shape by default.
        """
        sh = data.reduce(axes, self.keepdims,  # pylint: disable=E1101
                         dtype=numpy.int64)  # pylint: disable=E1101
        return (sh, )


if onnx_opset_version() >= 13:
    ReduceSum = ReduceSum_13
elif onnx_opset_version() >= 11:  # pragma: no cover
    ReduceSum = ReduceSum_11
else:  # pragma: no cover
    ReduceSum = ReduceSum_1



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ...onnx_tools.onnx2py_helper import guess_numpy_type_from_dtype
from ._op import OpRun
from ..shape_object import ShapeObject


class Concat(OpRun):

    atts = {'axis': 0}
    python_inputs = ['*inputs']

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Concat.atts,
                       **options)

    def _preprocess(self, a):
        if self.axis >= len(a.shape):
            new_shape = a.shape + (1, ) * (self.axis + 1 - len(a.shape))
            return a.reshape(new_shape)
        return a

    def _run(self, *args):  # pylint: disable=W0221
        targs = tuple(self._preprocess(a) for a in args)
        return (numpy.concatenate(targs, self.axis), )

    def _infer_shapes(self, *args):  # pylint: disable=W0221
        return (args[0].concat_columns(self.axis, *(args[1:])), )

    def _infer_types(self, *args):  # pylint: disable=W0221
        args = [guess_numpy_type_from_dtype(a) for a in args]
        res = (ShapeObject._infer_merged_type(*args, use_dtype=False), )
        return res

    def _infer_sizes(self, *args, **kwargs):
        res = self.run(*args, **kwargs)
        return (dict(temp=0), ) + res

    def to_python(self, inputs):
        return "import numpy", "return numpy.concatenate(inputs, axis=axis)"



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ..shape_object import ShapeObject
from .op_qlinear_conv_ import QLinearConvInt8, QLinearConvUInt8  # pylint: disable=E0611,E0401


class QLinearConv(OpRun):

    atts = {'auto_pad': 'NOTSET',
            'group': 1,
            'dilations': [],
            'kernel_shape': [],
            'pads': [],
            'strides': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=QLinearConv.atts,
                       **options)
        self._init()
        self._cstu8 = numpy.array([], dtype=numpy.uint8)
        self._csti8 = numpy.array([], dtype=numpy.int8)

    def _init(self):
        self.rtu8_ = QLinearConvUInt8()
        self.rti8_ = QLinearConvInt8()
        for rt in [self.rtu8_, self.rti8_]:
            rt.init(self.auto_pad,
                    numpy.array(self.dilations, dtype=numpy.int64),
                    self.group,
                    numpy.array(self.kernel_shape, dtype=numpy.int64),
                    numpy.array(self.pads, dtype=numpy.int64),
                    numpy.array(self.strides, dtype=numpy.int64))

    def _run(self, X, x_scale, x_zero_point, w, w_scale, w_zero_point,  # pylint: disable=W0221
             y_scale, y_zero_point, B=None):
        if X is None:
            raise ValueError(  # pragma: no cover
                "X cannot be None for operator %r, ONNX=%r" % (
                    type(self), self.onnx_node))
        if X.dtype == numpy.uint8:
            if B is None:
                b = self._cstu8
            else:
                b = B
            return (self.rtu8_.compute(
                X, x_scale, x_zero_point, w, w_scale, w_zero_point,  # pylint: disable=W0221
                y_scale, y_zero_point, b), )
        return (self.rti8_.compute(
            X, x_scale, x_zero_point, w, w_scale, w_zero_point,  # pylint: disable=W0221
            y_scale, y_zero_point, B or self._csti8), )

    def _infer_shapes(self, X, x_scale, x_zero_point, w, w_scale,  # pylint: disable=W0221
                      w_zero_point, y_scale, y_zero_point, B=None):

        return (ShapeObject(None, dtype=X.dtype), )

    def _infer_types(self, X, x_scale, x_zero_point, w, w_scale,  # pylint: disable=W0221
                     w_zero_point, y_scale, y_zero_point, B=None):

        return (X, )

    def _infer_sizes(self, *args, **kwargs):  # pylint: disable=W0221
        res = self.run(*args, **kwargs)
        return (dict(temp=0), ) + res

    def _infer_sizes(self, *args, **kwargs):  # pylint: disable=W0221
        res = self.run(*args, **kwargs)
        X = args[0]
        C = X.shape[1]
        kernel_size = numpy.prod(self.kernel_shape)
        kernel_dim = C / self.group * kernel_size
        temp = kernel_dim * res[0].size
        return (dict(temp=temp * X.dtype.itemsize), ) + res



"""
@file
@brief Helpers to validate python code.
"""
import pickle
import pprint
import numpy
from numpy.linalg import det as npy_det  # pylint: disable=E0611
from scipy.spatial.distance import cdist  # pylint: disable=E0611
from scipy.special import expit, erf  # pylint: disable=E0611
from scipy.linalg import solve  # pylint: disable=E0611
from ...tools.code_helper import make_callable


def _make_callable(fct, obj, code, gl, debug):
    """
    Same function as @see fn make_callable but deals with
    function which an undefined number of arguments.
    """
    def pyrt_Concat_(*inputs, axis=0):
        return numpy.concatenate(inputs, axis=axis)

    if fct == "pyrt_Concat":
        return pyrt_Concat_
    return make_callable(fct, obj, code, gl, debug)


def validate_python_inference(oinf, inputs, tolerance=0.):
    """
    Validates the code produced by method :meth:`to_python
    <mlprodict.onnxrt.onnx_inference_exports.OnnxInferenceExport.to_python>`.
    The function compiles and executes the code
    given as an argument and compares the results to
    what *oinf* returns. This function is mostly used for
    unit testing purpose but it is not robust enough
    to handle all cases.

    @param      oinf        @see cl OnnxInference
    @param      inputs      inputs as dictionary
    @param      tolerance   discrepencies must be below or equal to
                            this theshold

    The function fails if the expected output are not the same.
    """
    from ..ops_cpu.op_argmax import _argmax
    from ..ops_cpu.op_argmin import _argmin
    from ..ops_cpu.op_celu import _vcelu1

    cd = oinf.to_python()
    code = cd['onnx_pyrt_main.py']

    exp = oinf.run(inputs)
    if not isinstance(exp, dict):
        raise TypeError(  # pragma: no cover
            "exp is not a dictionary by '{}'.".format(type(exp)))
    if len(exp) == 0:
        raise ValueError(  # pragma: no cover
            "No result to compare.")
    inps = ['{0}={0}'.format(k) for k in sorted(inputs)]
    code += "\n".join(['', '', 'opi = OnnxPythonInference()',
                       'res = opi.run(%s)' % ', '.join(inps)])

    cp = compile(code, "<string>", mode='exec')
    pyrt_fcts = [_ for _ in cp.co_names if _.startswith("pyrt_")]
    fcts_local = {}

    gl = {'numpy': numpy, 'pickle': pickle,
          'expit': expit, 'erf': erf, 'cdist': cdist,
          '_argmax': _argmax, '_argmin': _argmin,
          '_vcelu1': _vcelu1, 'solve': solve,
          'npy_det': npy_det, 'ndarray': numpy.ndarray}

    for fct in pyrt_fcts:
        for obj in cp.co_consts:
            if isinstance(obj, str):
                continue
            sobj = str(obj)
            if '<string>' in sobj and fct in sobj:
                fcts_local[fct] = _make_callable(fct, obj, code, gl, False)

    gl.update(fcts_local)
    loc = inputs
    try:
        exec(cp, gl, loc)  # pylint: disable=W0122
    except (NameError, TypeError, SyntaxError) as e:  # pragma: no cover
        raise RuntimeError(
            "Unable to execute code\n-----\n{}".format(code)) from e

    got = loc['res']
    keys = list(sorted(exp))
    if isinstance(got, numpy.ndarray) and len(keys) == 1:
        got = {keys[0]: got}

    if not isinstance(got, dict):
        raise TypeError(  # pragma: no cover
            "got is not a dictionary by '{}'\n--\n{}\n---\n{}.".format(
                type(got), dir(got), pprint.pformat(str(loc))))
    if len(got) != len(exp):
        raise RuntimeError(  # pragma: no cover
            "Different number of results.\nexp: {}\ngot: {}".format(
                ", ".join(sorted(exp)), ", ".join(sorted(got))))

    if keys != list(sorted(got)):
        raise RuntimeError(  # pragma: no cover
            "Different result names.\nexp: {}\ngot: {}".format(
                ", ".join(sorted(exp)), ", ".join(sorted(got))))

    for k in keys:
        e = exp[k]
        g = got[k]
        if isinstance(e, numpy.ndarray):
            if e.shape != g.shape:
                raise ValueError(  # pragma: no cover
                    "Shapes are different {} != {}\n---\n{}\n{}.".format(
                        e.shape, g.shape, e, g))
            diff = 0
            for a, b in zip(e.ravel(), g.ravel()):
                if a == b:
                    continue
                if (isinstance(a, float) and isinstance(b, float) and
                        numpy.isnan(a) and numpy.isnan(b)):
                    continue  # pragma: no cover
                diff = max(diff, abs(a - b))
            if diff > tolerance:
                raise ValueError(  # pragma: no cover
                    "Values are different (max diff={}>{})\n--EXP--\n{}\n--GOT--"
                    "\n{}\n--\n{}".format(diff, tolerance, e, g, code))
        else:
            raise NotImplementedError(  # pragma: no cover
                "Unable to compare values of type '{}'.".format(type(e)))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ._op import OpRun


class SequenceInsert(OpRun):

    atts = {'axis': 0, 'new_axis': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       atts=SequenceInsert.atts, **options)

    def _run(self, S, T, ind=None):  # pylint: disable=W0221
        S = S.copy()
        if ind is not None:
            S.insert(ind[0], T)
        else:
            S.append(T)
        return (S, )

    def _infer_shapes(self, S, T, ind=None):  # pylint: disable=W0221
        return (S, )

    def _infer_types(self, S, T, ind=None):  # pylint: disable=W0221
        return (S, )

    def _infer_sizes(self, *args):  # pylint: disable=W0221
        res = self.run(*args)
        return (dict(temp=0), ) + res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ..shape_object import ShapeObject
from ._op import OpRun


class Range(OpRun):

    atts = {}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Range.atts,
                       **options)

    def _run(self, starts, ends, steps):  # pylint: disable=W0221
        return (numpy.arange(starts, ends, steps).astype(starts.dtype), )

    def _infer_shapes(self, starts, ends, steps):  # pylint: disable=W0221
        return (ShapeObject(None, starts.dtype), )

    def _infer_types(self, starts, ends, steps):  # pylint: disable=W0221
        return (starts, )

    def _infer_sizes(self, *args, **kwargs):  # pylint: disable=W0221
        res = self.run(*args, **kwargs)
        return (dict(temp=0), ) + res



"""
@file
@brief Modified converter from
`LightGbm.py <https://github.com/onnx/onnxmltools/blob/master/onnxmltools/convert/
lightgbm/operator_converters/LightGbm.py>`_.
"""
from collections import Counter
import copy
import numbers
import numpy
from skl2onnx.common._apply_operation import apply_div, apply_reshape, apply_sub  # pylint: disable=E0611
from skl2onnx.common.tree_ensemble import get_default_tree_classifier_attribute_pairs
from skl2onnx.proto import onnx_proto
from skl2onnx.common.shape_calculator import (
    calculate_linear_regressor_output_shapes,
    calculate_linear_classifier_output_shapes)
from skl2onnx.common.data_types import guess_numpy_type


def calculate_lightgbm_output_shapes(operator):
    """
    Shape calculator for LightGBM Booster
    (see :epkg:`lightgbm`).
    """
    op = operator.raw_operator
    if hasattr(op, "_model_dict"):
        objective = op._model_dict['objective']
    else:
        objective = op.objective_
    if objective.startswith('binary') or objective.startswith('multiclass'):
        return calculate_linear_classifier_output_shapes(operator)
    if objective.startswith('regression'):  # pragma: no cover
        return calculate_linear_regressor_output_shapes(operator)
    raise NotImplementedError(  # pragma: no cover
        "Objective '{}' is not implemented yet.".format(objective))


def _translate_split_criterion(criterion):
    # If the criterion is true, LightGBM use the left child. Otherwise, right child is selected.
    if criterion == '<=':
        return 'BRANCH_LEQ'
    if criterion == '<':  # pragma: no cover
        return 'BRANCH_LT'
    if criterion == '>=':  # pragma: no cover
        return 'BRANCH_GTE'
    if criterion == '>':  # pragma: no cover
        return 'BRANCH_GT'
    if criterion == '==':  # pragma: no cover
        return 'BRANCH_EQ'
    if criterion == '!=':  # pragma: no cover
        return 'BRANCH_NEQ'
    raise ValueError(  # pragma: no cover
        'Unsupported splitting criterion: %s. Only <=, <, >=, and > are allowed.')


def _create_node_id(node_id_pool):
    i = 0
    while i in node_id_pool:
        i += 1
    node_id_pool.add(i)
    return i


def _parse_tree_structure(tree_id, class_id, learning_rate, tree_structure, attrs):
    """
    The pool of all nodes' indexes created when parsing a single tree.
    Different tree use different pools.
    """
    node_id_pool = set()
    node_pyid_pool = dict()

    node_id = _create_node_id(node_id_pool)
    node_pyid_pool[id(tree_structure)] = node_id

    # The root node is a leaf node.
    if 'left_child' not in tree_structure or 'right_child' not in tree_structure:
        _parse_node(tree_id, class_id, node_id, node_id_pool, node_pyid_pool,
                    learning_rate, tree_structure, attrs)
        return

    left_pyid = id(tree_structure['left_child'])
    right_pyid = id(tree_structure['right_child'])

    if left_pyid in node_pyid_pool:
        left_id = node_pyid_pool[left_pyid]
        left_parse = False
    else:
        left_id = _create_node_id(node_id_pool)
        node_pyid_pool[left_pyid] = left_id
        left_parse = True

    if right_pyid in node_pyid_pool:
        right_id = node_pyid_pool[right_pyid]
        right_parse = False
    else:
        right_id = _create_node_id(node_id_pool)
        node_pyid_pool[right_pyid] = right_id
        right_parse = True

    attrs['nodes_treeids'].append(tree_id)
    attrs['nodes_nodeids'].append(node_id)

    attrs['nodes_featureids'].append(tree_structure['split_feature'])
    attrs['nodes_modes'].append(
        _translate_split_criterion(tree_structure['decision_type']))
    if isinstance(tree_structure['threshold'], str):
        try:  # pragma: no cover
            attrs['nodes_values'].append(  # pragma: no cover
                float(tree_structure['threshold']))
        except ValueError as e:  # pragma: no cover
            import pprint
            text = pprint.pformat(tree_structure)
            if len(text) > 99999:
                text = text[:99999] + "\n..."
            raise TypeError("threshold must be a number not '{}'"
                            "\n{}".format(tree_structure['threshold'], text)) from e
    else:
        attrs['nodes_values'].append(tree_structure['threshold'])

    # Assume left is the true branch and right is the false branch
    attrs['nodes_truenodeids'].append(left_id)
    attrs['nodes_falsenodeids'].append(right_id)
    if tree_structure['default_left']:
        attrs['nodes_missing_value_tracks_true'].append(1)
    else:
        attrs['nodes_missing_value_tracks_true'].append(0)
    attrs['nodes_hitrates'].append(1.)
    if left_parse:
        _parse_node(tree_id, class_id, left_id, node_id_pool, node_pyid_pool,
                    learning_rate, tree_structure['left_child'], attrs)
    if right_parse:
        _parse_node(tree_id, class_id, right_id, node_id_pool, node_pyid_pool,
                    learning_rate, tree_structure['right_child'], attrs)


def _parse_node(tree_id, class_id, node_id, node_id_pool, node_pyid_pool,
                learning_rate, node, attrs):
    """
    Parses nodes.
    """
    if (hasattr(node, 'left_child') and hasattr(node, 'right_child')) or \
            ('left_child' in node and 'right_child' in node):

        left_pyid = id(node['left_child'])
        right_pyid = id(node['right_child'])

        if left_pyid in node_pyid_pool:
            left_id = node_pyid_pool[left_pyid]
            left_parse = False
        else:
            left_id = _create_node_id(node_id_pool)
            node_pyid_pool[left_pyid] = left_id
            left_parse = True

        if right_pyid in node_pyid_pool:
            right_id = node_pyid_pool[right_pyid]
            right_parse = False
        else:
            right_id = _create_node_id(node_id_pool)
            node_pyid_pool[right_pyid] = right_id
            right_parse = True

        attrs['nodes_treeids'].append(tree_id)
        attrs['nodes_nodeids'].append(node_id)

        attrs['nodes_featureids'].append(node['split_feature'])
        attrs['nodes_modes'].append(
            _translate_split_criterion(node['decision_type']))
        if isinstance(node['threshold'], str):
            try:  # pragma: no cover
                attrs['nodes_values'].append(  # pragma: no cover
                    float(node['threshold']))
            except ValueError as e:  # pragma: no cover
                import pprint
                text = pprint.pformat(node)
                if len(text) > 99999:
                    text = text[:99999] + "\n..."
                raise TypeError("threshold must be a number not '{}'"
                                "\n{}".format(node['threshold'], text)) from e
        else:
            attrs['nodes_values'].append(node['threshold'])

        # Assume left is the true branch and right is the false branch
        attrs['nodes_truenodeids'].append(left_id)
        attrs['nodes_falsenodeids'].append(right_id)
        if node['default_left']:
            attrs['nodes_missing_value_tracks_true'].append(1)
        else:
            attrs['nodes_missing_value_tracks_true'].append(0)
        attrs['nodes_hitrates'].append(1.)

        # Recursively dive into the child nodes
        if left_parse:
            _parse_node(tree_id, class_id, left_id, node_id_pool, node_pyid_pool,
                        learning_rate, node['left_child'], attrs)
        if right_parse:
            _parse_node(tree_id, class_id, right_id, node_id_pool, node_pyid_pool,
                        learning_rate, node['right_child'], attrs)
    elif hasattr(node, 'left_child') or hasattr(node, 'right_child'):
        raise ValueError('Need two branches')  # pragma: no cover
    else:
        # Node attributes
        attrs['nodes_treeids'].append(tree_id)
        attrs['nodes_nodeids'].append(node_id)
        attrs['nodes_featureids'].append(0)
        attrs['nodes_modes'].append('LEAF')
        # Leaf node has no threshold. A zero is appended but it will never be used.
        attrs['nodes_values'].append(0.)
        # Leaf node has no child. A zero is appended but it will never be used.
        attrs['nodes_truenodeids'].append(0)
        # Leaf node has no child. A zero is appended but it will never be used.
        attrs['nodes_falsenodeids'].append(0)
        # Leaf node has no split function. A zero is appended but it will never be used.
        attrs['nodes_missing_value_tracks_true'].append(0)
        attrs['nodes_hitrates'].append(1.)

        # Leaf attributes
        attrs['class_treeids'].append(tree_id)
        attrs['class_nodeids'].append(node_id)
        attrs['class_ids'].append(class_id)
        attrs['class_weights'].append(
            float(node['leaf_value']) * learning_rate)


def convert_lightgbm(scope, operator, container):
    """
    This converters reuses the code from
    `LightGbm.py <https://github.com/onnx/onnxmltools/blob/master/onnxmltools/convert/
    lightgbm/operator_converters/LightGbm.py>`_ and makes
    some modifications. It implements converters
    for models in :epkg:`lightgbm`.
    """
    gbm_model = operator.raw_operator
    gbm_text = gbm_model.booster_.dump_model()
    modify_tree_for_rule_in_set(gbm_text, use_float=True)

    attrs = get_default_tree_classifier_attribute_pairs()
    attrs['name'] = operator.full_name

    # Create different attributes for classifier and regressor, respectively
    if gbm_text['objective'].startswith('binary'):
        n_classes = 1
        attrs['post_transform'] = 'LOGISTIC'
    elif gbm_text['objective'].startswith('multiclass'):
        n_classes = gbm_text['num_class']
        attrs['post_transform'] = 'SOFTMAX'
    elif gbm_text['objective'].startswith('regression'):
        n_classes = 1  # Regressor has only one output variable
        attrs['post_transform'] = 'NONE'
        attrs['n_targets'] = n_classes
    else:
        raise RuntimeError(  # pragma: no cover
            "LightGBM objective should be cleaned already not '{}'.".format(
                gbm_text['objective']))

    # Use the same algorithm to parse the tree
    for i, tree in enumerate(gbm_text['tree_info']):
        tree_id = i
        class_id = tree_id % n_classes
        # tree['shrinkage'] --> LightGbm provides figures with it already.
        learning_rate = 1.
        _parse_tree_structure(
            tree_id, class_id, learning_rate, tree['tree_structure'], attrs)

    # Sort nodes_* attributes. For one tree, its node indexes should appear in an ascent order in nodes_nodeids. Nodes
    # from a tree with a smaller tree index should appear before trees with larger indexes in nodes_nodeids.
    node_numbers_per_tree = Counter(attrs['nodes_treeids'])
    tree_number = len(node_numbers_per_tree.keys())
    accumulated_node_numbers = [0] * tree_number
    for i in range(1, tree_number):
        accumulated_node_numbers[i] = (accumulated_node_numbers[i - 1] +
                                       node_numbers_per_tree[i - 1])
    global_node_indexes = []
    for i in range(len(attrs['nodes_nodeids'])):
        tree_id = attrs['nodes_treeids'][i]
        node_id = attrs['nodes_nodeids'][i]
        global_node_indexes.append(accumulated_node_numbers[tree_id] + node_id)
    for k, v in attrs.items():
        if k.startswith('nodes_'):
            merged_indexes = zip(copy.deepcopy(global_node_indexes), v)
            sorted_list = [pair[1]
                           for pair in sorted(merged_indexes, key=lambda x: x[0])]
            attrs[k] = sorted_list

    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32

    # Create ONNX object
    if (gbm_text['objective'].startswith('binary') or
            gbm_text['objective'].startswith('multiclass')):
        # Prepare label information for both of TreeEnsembleClassifier
        # and ZipMap
        class_type = onnx_proto.TensorProto.STRING  # pylint: disable=E1101
        zipmap_attrs = {'name': scope.get_unique_variable_name('ZipMap')}
        if all(isinstance(i, (numbers.Real, bool, numpy.bool_))
               for i in gbm_model.classes_):
            class_type = onnx_proto.TensorProto.INT64  # pylint: disable=E1101
            class_labels = [int(i) for i in gbm_model.classes_]
            attrs['classlabels_int64s'] = class_labels
            zipmap_attrs['classlabels_int64s'] = class_labels
        elif all(isinstance(i, str) for i in gbm_model.classes_):
            class_labels = [str(i) for i in gbm_model.classes_]
            attrs['classlabels_strings'] = class_labels
            zipmap_attrs['classlabels_strings'] = class_labels
        else:
            raise ValueError(  # pragma: no cover
                'Only string and integer class labels are allowed')

        # Create tree classifier
        probability_tensor_name = scope.get_unique_variable_name(
            'probability_tensor')
        label_tensor_name = scope.get_unique_variable_name('label_tensor')

        if dtype == numpy.float64:
            container.add_node('TreeEnsembleClassifierDouble', operator.input_full_names,
                               [label_tensor_name, probability_tensor_name],
                               op_domain='mlprodict', **attrs)
        else:
            container.add_node('TreeEnsembleClassifier', operator.input_full_names,
                               [label_tensor_name, probability_tensor_name],
                               op_domain='ai.onnx.ml', **attrs)

        prob_tensor = probability_tensor_name

        if gbm_model.boosting_type == 'rf':
            col_index_name = scope.get_unique_variable_name('col_index')
            first_col_name = scope.get_unique_variable_name('first_col')
            zeroth_col_name = scope.get_unique_variable_name('zeroth_col')
            denominator_name = scope.get_unique_variable_name('denominator')
            modified_first_col_name = scope.get_unique_variable_name(
                'modified_first_col')
            unit_float_tensor_name = scope.get_unique_variable_name(
                'unit_float_tensor')
            merged_prob_name = scope.get_unique_variable_name('merged_prob')
            predicted_label_name = scope.get_unique_variable_name(
                'predicted_label')
            classes_name = scope.get_unique_variable_name('classes')
            final_label_name = scope.get_unique_variable_name('final_label')

            container.add_initializer(
                col_index_name, onnx_proto.TensorProto.INT64, [], [1])  # pylint: disable=E1101
            container.add_initializer(
                unit_float_tensor_name, onnx_proto.TensorProto.FLOAT, [], [1.0])  # pylint: disable=E1101
            container.add_initializer(
                denominator_name, onnx_proto.TensorProto.FLOAT, [], [100.0])  # pylint: disable=E1101
            container.add_initializer(classes_name, class_type,
                                      [len(class_labels)], class_labels)

            container.add_node('ArrayFeatureExtractor', [probability_tensor_name, col_index_name],
                               first_col_name, name=scope.get_unique_operator_name(
                                   'ArrayFeatureExtractor'),
                               op_domain='ai.onnx.ml')
            apply_div(scope, [first_col_name, denominator_name],
                      modified_first_col_name, container, broadcast=1)
            apply_sub(scope, [unit_float_tensor_name, modified_first_col_name],
                      zeroth_col_name, container, broadcast=1)
            container.add_node('Concat', [zeroth_col_name, modified_first_col_name],
                               merged_prob_name, name=scope.get_unique_operator_name('Concat'), axis=1)
            container.add_node('ArgMax', merged_prob_name,
                               predicted_label_name, name=scope.get_unique_operator_name('ArgMax'), axis=1)
            container.add_node('ArrayFeatureExtractor', [classes_name, predicted_label_name], final_label_name,
                               name=scope.get_unique_operator_name('ArrayFeatureExtractor'), op_domain='ai.onnx.ml')
            apply_reshape(scope, final_label_name,
                          operator.outputs[0].full_name, container, desired_shape=[-1, ])
            prob_tensor = merged_prob_name
        else:
            container.add_node('Identity', label_tensor_name,
                               operator.outputs[0].full_name,
                               name=scope.get_unique_operator_name('Identity'))

        # Convert probability tensor to probability map
        # (keys are labels while values are the associated probabilities)
        container.add_node('Identity', prob_tensor,
                           operator.outputs[1].full_name)
    else:
        # Create tree regressor
        output_name = scope.get_unique_variable_name('output')

        keys_to_be_renamed = list(
            k for k in attrs if k.startswith('class_'))

        for k in keys_to_be_renamed:
            # Rename class_* attribute to target_* because TreeEnsebmleClassifier
            # and TreeEnsembleClassifier have different ONNX attributes
            attrs['target' + k[5:]] = copy.deepcopy(attrs[k])
            del attrs[k]
        if dtype == numpy.float64:
            container.add_node('TreeEnsembleRegressorDouble', operator.input_full_names,
                               output_name, op_domain='mlprodict', **attrs)
        else:
            container.add_node('TreeEnsembleRegressor', operator.input_full_names,
                               output_name, op_domain='ai.onnx.ml', **attrs)
        if gbm_model.boosting_type == 'rf':
            denominator_name = scope.get_unique_variable_name('denominator')

            container.add_initializer(
                denominator_name, onnx_proto.TensorProto.FLOAT, [], [100.0])  # pylint: disable=E1101

            apply_div(scope, [output_name, denominator_name],
                      operator.output_full_names, container, broadcast=1)
        else:
            container.add_node('Identity', output_name,
                               operator.output_full_names,
                               name=scope.get_unique_operator_name('Identity'))


def modify_tree_for_rule_in_set(gbm, use_float=False):  # pylint: disable=R1710
    """
    LightGBM produces sometimes a tree with a node set
    to use rule ``==`` to a set of values (= in set),
    the values are separated by ``||``.
    This function unfold theses nodes. A child looks
    like the following:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import pprint
        from mlprodict.onnx_conv.operator_converters.conv_lightgbm import modify_tree_for_rule_in_set

        tree = {'decision_type': '==',
                'default_left': True,
                'internal_count': 6805,
                'internal_value': 0.117558,
                'left_child': {'leaf_count': 4293,
                               'leaf_index': 18,
                               'leaf_value': 0.003519117642745049},
                'missing_type': 'None',
                'right_child': {'leaf_count': 2512,
                                'leaf_index': 25,
                                'leaf_value': 0.012305307958365394},
                'split_feature': 24,
                'split_gain': 12.233599662780762,
                'split_index': 24,
                'threshold': '10||12||13'}

        modify_tree_for_rule_in_set(tree)

        pprint.pprint(tree)
    """
    if 'tree_info' in gbm:
        for tree in gbm['tree_info']:
            modify_tree_for_rule_in_set(tree, use_float=use_float)
        return

    if 'tree_structure' in gbm:
        modify_tree_for_rule_in_set(gbm['tree_structure'], use_float=use_float)
        return

    if 'decision_type' not in gbm:
        return

    def recursive_call(this):
        if 'left_child' in this:
            modify_tree_for_rule_in_set(
                this['left_child'], use_float=use_float)
        if 'right_child' in this:
            modify_tree_for_rule_in_set(
                this['right_child'], use_float=use_float)

    def str2number(val):
        if use_float:
            return float(val)
        else:
            try:
                return int(val)
            except ValueError:  # pragma: no cover
                return float(val)

    dec = gbm['decision_type']
    if dec != '==':
        return recursive_call(gbm)

    th = gbm['threshold']
    if not isinstance(th, str) or '||' not in th:
        return recursive_call(gbm)

    pos = th.index('||')
    th1 = str2number(th[:pos])

    rest = th[pos + 2:]
    if '||' not in rest:
        rest = str2number(rest)

    gbm['threshold'] = th1
    new_node = gbm.copy()
    gbm['right_child'] = new_node
    new_node['threshold'] = rest
    return recursive_call(gbm)
